from unittest.mock import Mock

import pytest

from decoda.well_known import *


def test_location_fmt():
    assert location_fmt(0x00) == "1st axle, 1st tire"
    assert location_fmt(0xFF) == "Not available"
    assert location_fmt(0x23) == "3rd axle, 4th tire"
    assert location_fmt(0x11) == "2nd axle, 2nd tire"


def test_refers_to_spn():
    try:
        assert refers_to_spn(0) == "SPN 0"
        assert (
            refers_to_spn(22)
            == "SPN 22 - Engine Extended Crankcase Blow-by Pressure"
        )
    except FileNotFoundError:
        pytest.skip("Unable to find a spec file")


class TestConditionallyApplies:
    def test_it_raises_error_if_conditional_on_spn_not_decoded(self):
        with pytest.raises(ValueError) as e:
            conditionally_applies(
                0x00,
                already_decoded=[Mock(id=9999)],
                custom_args={"conditional_on_spn": 1234, "applies_if": 4321},
            )
        assert str(e.value) == "conditional on SPN 1234, but not found"

    @pytest.mark.parametrize(
        "custom_args",
        [
            {},
            {"conditional_on_spn": 1234},
            {"applies_if": 9999},
        ],
    )
    def test_it_raises_bad_config_if_missing_config(self, custom_args):
        with pytest.raises(ValueError) as e:
            conditionally_applies(
                0x00, already_decoded=[Mock(id=9999)], custom_args=custom_args
            )
        assert str(e.value) == "bad config"

    def test_it_raises_error_if_no_decoded_values_provided(self):
        with pytest.raises(ValueError) as e:
            conditionally_applies(
                0x00,
                already_decoded=[],
                custom_args={"conditional_on_spn": 1234, "applies_if": 4321},
            )
        assert str(e.value) == "no values decoded"

    def test_it_returns_none_if_decoded_value_does_not_match_applies_if(self):
        value = conditionally_applies(
            0x00,
            already_decoded=[Mock(id=9999)],
            custom_args={"conditional_on_spn": 9999, "applies_if": 4321},
        )
        assert value is None

    def test_it_returns_value_from_non_custom_alternative_if_matches_applies_if(
        self,
    ):
        non_custom_alternative = Mock()
        non_custom_alternative.decode_from_raw.return_value = [
            {"value": "the return value"}
        ]
        value = conditionally_applies(
            0x00,
            already_decoded=[Mock(id=9999, value="decoded value")],
            custom_args={
                "conditional_on_spn": 9999,
                "applies_if": "decoded value",
            },
            non_custom_alternative=non_custom_alternative,
        )
        assert value == "the return value"



import argparse
import json
import sys

import xlrd2 as xlrd

from .json_from_da import extract_pgns, extract_spns, secure_open_workbook


def filter_source_doc(item):
    return "11783" in item.get("source_document")


def spec_from_workbook(wb: xlrd.Book):
    return {
        "PGNs": [pgn for pgn in extract_pgns(wb) if filter_source_doc(pgn)],
        "SPNs": [spn for spn in extract_spns(wb) if filter_source_doc(spn)],
    }


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "isobus_pgns_spns_xlsx",
        type=str,
        help="The SPGs and PGNs XLSX workbook from isobus.net",
    )
    parser.add_argument(
        "json_file",
        nargs="?",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="the json file to use as output",
    )
    parser.add_argument("--pretty", default=False, action="store_true")
    args = parser.parse_args()

    print(args.isobus_pgns_spns_xlsx)
    wb = secure_open_workbook(
        filename=args.isobus_pgns_spns_xlsx, on_demand=True
    )
    spec = spec_from_workbook(wb)
    json.dump(spec, args.json_file, indent=2 if args.pretty else None)


if __name__ == "__main__":
    main()



import argparse
import json
import re
import sys

from decoda.spec_loader import spn_from_dict


def remove_pgns(pgns):
    result = []
    for pgn in pgns:
        spns_list = pgn["spns"]
        if len(spns_list) > 0 and all(
            bool(spn.get("start_pos")) for spn in spns_list
        ):
            # we want this, even if it appears to be invalid....we should fix
            # it
            result.append(pgn)
        elif pgn.get("length") != "":
            result.append(pgn)

    return result


def remove_spns(spns):
    result = []
    for spn in spns:
        try:
            spn_from_dict(spn)
            result.append(spn)
        except (ValueError, KeyError):
            pass
    return result


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "input_file",
        nargs="?",
        type=argparse.FileType("r"),
        default=sys.stdout,
        help="the json file to use as input",
    )
    parser.add_argument(
        "output_file",
        nargs="?",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="the json file to use as output",
    )
    parser.add_argument("--pretty", default=False, action="store_true")
    args = parser.parse_args()

    spec = json.load(args.input_file)
    spec["PGNs"] = remove_pgns(spec["PGNs"])
    spec["SPNs"] = remove_spns(spec["SPNs"])
    json.dump(spec, args.output_file, indent=2 if args.pretty else None)


if __name__ == "__main__":
    main()



import argparse
import json
import pathlib
import sys


def load_corrections(path, key):
    result = []
    for filename in path.glob("*.json"):
        with open(filename) as f:
            corrections_list = json.load(f)[key]
            result.append({obj["id"]: obj for obj in corrections_list})
    return result


# items is a list of dicts, where each item is an SPN or PGN dict
# updates is a list of dicts, where each update is a dictionary of "id" to the
# modifications needed
def treat_list_with_updates(items, updates):
    result = []

    for item in items:
        # Apply updates from all update dicts in order
        for update in updates:
            if item["id"] in update:
                item.update(update.get(item["id"], {}))
                del update[item["id"]]

        result.append(item)

    # Add any "updates" that were not applied. This allows the corrections
    # files to specify new items entirely
    for update in updates:
        result.extend(update.values())

    result = sorted(result, key=lambda obj: obj["id"])
    return result


def replace_bad_resolutions(spns):
    result = []
    for spn in spns:
        resolution = spn.get("resolution")
        if resolution and type(resolution) is str:
            print(
                f"Converting resolution for SPN: {spn['id']} to 1, from: {spn['resolution']}"
            )
            spn["resolution"] = 1
        result.append(spn)
    return result


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--corrections_path",
        type=pathlib.Path,
        default="./corrections",
        help="the path to corrections to apply",
    )
    parser.add_argument(
        "input_file",
        nargs="?",
        type=argparse.FileType("r"),
        default=sys.stdout,
        help="the json file to use as input",
    )
    parser.add_argument(
        "output_file",
        nargs="?",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="the json file to use as output",
    )
    parser.add_argument("--pretty", default=False, action="store_true")
    args = parser.parse_args()

    pgn_corrections = load_corrections(args.corrections_path, "PGNs")
    spn_corrections = load_corrections(args.corrections_path, "SPNs")

    spec = json.load(args.input_file)
    spec["PGNs"] = treat_list_with_updates(spec["PGNs"], pgn_corrections)
    spec["SPNs"] = treat_list_with_updates(spec["SPNs"], spn_corrections)
    spec["SPNs"] = replace_bad_resolutions(spec["SPNs"])

    json.dump(spec, args.output_file, indent=2 if args.pretty else None)


if __name__ == "__main__":
    main()



from .main import *
from .spec_loader import *
from .transport import *

__version__ = "0.2.6"

__title__ = "decoda"
__description__ = "Decoder for J1939"
__url__ = "https://www.decoda.cc/"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Andrew Dodd"
__email__ = "andrew.john.dodd@gmail.com"

__license__ = "MIT"
__copyright__ = "Copyright (c) Andrew Dodd"



import json

import pytest

from decoda import *


def test_mask():
    assert mask(0) == 0
    assert mask(1) == 0b1
    assert mask(2) == 0b11
    assert mask(3) == 0b111
    assert mask(9) == 0b111111111
    assert mask(50) == 0b11111111111111111111111111111111111111111111111111


class TestRegularPackedPGNs:
    def test_it_unpacks_PGN_0_Torque_Speed_Control_1(self, pgn_0):
        """This is the most basic PGN...which includes encoded values and regular values"""
        sut = pgn_0
        assert sut.id == 0
        assert sut.name == "Torque/Speed Control 1"

        inp = (0x123456781AB7DEF0).to_bytes(8, "big")
        decoded = sut.decode(inp)

        assert len(decoded) == 10
        # 0x12 - 0b00010010
        assert (
            decoded[0].value
            == 'Torque control - Control torque to the included "desired torque" value.'
        )
        assert (
            decoded[1].value
            == "Transient Optimized for driveline disengaged and non-lockup conditions"
        )  # yapf: disable
        assert decoded[2].value == "High priority"
        # 0x5634 - 22068 == 2758.5
        assert decoded[3].value == 2758.5
        # 0x78 = 120
        assert decoded[4].value == -5

        # 0x1a - 0b00011010
        assert decoded[5].value == "500 ms transmission rate"
        assert decoded[6].value == "P4 = Road Speed Governor"

        assert decoded[7].raw == 7
        assert decoded[7].value == "0.875 %"

        # 0xde - not decoded
        # 0xf0 -
        assert decoded[8].value == 0
        assert decoded[9].value == 15

        for d in decoded:
            print("{:<45}: {}".format(d.name, d.display_value))
        # assert False

    def test_it_unpacks_PGN_61444_Electronic_Engine_Controller_1(
        self, pgn_61444
    ):
        """This is a common PGN"""
        sut = pgn_61444
        assert sut.id == 61444
        assert sut.name == "Electronic Engine Controller 1"

        inp = (0x123456781ABCDEF0).to_bytes(8, "big")
        decoded = sut.decode(inp)

        assert len(decoded) == 8
        assert decoded[0].value == "Cruise control"
        assert decoded[1].value == "0.125 %"
        assert decoded[2].value == -73
        assert decoded[3].value == -39
        assert decoded[4].value == 847.0
        assert decoded[5].value == 188
        assert decoded[6].value == "error"
        assert decoded[7].value == 115

        for d in decoded:
            print("{:<45}: {}".format(d.name, d.display_value))

        inp = (0xFFFF9100FDFFFFFF).to_bytes(8, "big")
        decoded = sut.decode(inp)

        assert len(decoded) == 8
        assert decoded[0].value == "Not available"
        assert decoded[1].value == "Not available"
        assert decoded[2].value == "Not available"
        assert decoded[3].value == 20
        assert decoded[4].value == "Parameter specific indicator"
        assert decoded[5].value == "Not available"
        assert decoded[6].value == "not available"
        assert decoded[7].value == "Not available"
        for d in decoded:
            print("{:<45}: {}".format(d.name, d.display_value))

        # assert False

    def test_it_unpacks_PGN_55552_MemoryAccessRequest(self, pgn_55552):
        """This PGN has a non-contigious SPN 1640"""
        sut = pgn_55552
        assert sut.id == 55552
        assert sut.name == "Memory Access Request"

        inp = (0x123456789ABCDEF0).to_bytes(8, "big")
        decoded = sut.decode(inp)

        assert len(decoded) == 6
        # 0x3412 = 0b0011 0100 0001 0010
        #            iiix ^^^x iiii iiii
        #              >> 0b 00100010010
        assert decoded[0].value == 0b00100010010
        assert decoded[1].raw == 2
        assert str(decoded[1]) == "SPN:1642: Command = Write"

        assert decoded[2].value == "Direct Spatial Addressing"
        # 0x9a7856 = 10123350
        assert decoded[3].value == 10123350
        # 0xbc = 188
        assert decoded[4].value == 188
        # 0xf0de = 61662
        assert decoded[5].value == 61662


class TestPGNsWithVariableText:
    def test_it_unpacks_with_one_variable_length(self, pgn_43008):
        sut = pgn_43008
        assert sut.id == 43008
        # NB unused byte here -------+
        #                            V
        inp = (
            bytes([3, 9, 0x12])
            + bytes("Abc123", "ascii")
            + bytes([0])
            + bytes("xyz", "ascii")
        )
        decoded = sut.decode(inp)

        assert decoded[0].raw == 3
        assert (
            decoded[0].value
            == "Overwrite display - The presently displayed information is to be overwritten with the transmitted information"
        )
        assert decoded[1].value == 0x12
        assert decoded[2].value == "Abc123"

    def test_it_unpacks_multiple_variable_length(self, pgn_64965):
        sut = pgn_64965
        assert sut.id == 64965
        decoded = sut.decode(bytes("hi*how*are**you*", "ascii"))

        assert decoded[0].value == "hi"
        assert decoded[1].value == "how"
        assert decoded[2].value == "are"
        assert decoded[3].value == ""
        assert decoded[4].value == "you"

    def test_it_unpacks_with_variable_text_as_well_as_regular(self, pgn_64958):
        sut = pgn_64958
        assert sut.id == 64958
        # TODO - This data is incorrect. I don't think the strings should be
        # delimited...instead it should be based on the lengths in 3071, 3072,
        # 3073
        decoded = sut.decode(bytes("\1\6\4\5Route*Run*Block", "ascii"))

        assert decoded[0].value == 1
        assert decoded[1].value == 6
        assert decoded[2].value == 4
        assert decoded[3].value == 5
        assert decoded[4].value == "Route*"
        assert decoded[5].value == "Run*"
        assert decoded[6].value == "Block"

    def test_it_unpacks_with_comma_specs(self, pgn_10240):
        sut = pgn_10240
        assert sut.id == 10240

        inp = (0x123456781AB7DEF0).to_bytes(8, "big")
        decoded = sut.decode(inp)
        assert decoded[0].raw == 0x12
        assert decoded[1].raw == 0x04
        assert decoded[2].raw == 0x563
        assert decoded[3].raw == 0xA78


class TestPGNsWithRepeatableSPNs:
    def test_it_unpacks_PGN_65226_ActiveDiagnosticTroubleCodes(
        self, pgn_65226: PGN
    ):
        sut = pgn_65226
        assert sut.id == 65226
        assert sut.name == "Active Diagnostic Trouble Codes"

        #         1 2 3 4 5 6 7 8 9 10 ...
        inp = (0x5555B8041000BBBB6000).to_bytes(10, "big")
        decoded = sut.decode(inp)

        assert decoded[0].value == "Lamp On"
        assert decoded[1].value == "Lamp On"
        assert decoded[2].value == "Lamp On"
        assert decoded[3].value == "Lamp On"
        assert (
            decoded[4].value == "Fast Flash (2 Hz or faster, 50% duty cycle)"
        )
        assert (
            decoded[5].value == "Fast Flash (2 Hz or faster, 50% duty cycle)"
        )
        assert (
            decoded[6].value == "Fast Flash (2 Hz or faster, 50% duty cycle)"
        )
        assert (
            decoded[7].value == "Fast Flash (2 Hz or faster, 50% duty cycle)"
        )

        # First SPN
        assert decoded[8].raw == 0x004B8
        assert (
            decoded[8].value == "SPN 1208 - Engine Oil Filter Intake Pressure"
        )
        assert (
            decoded[8].display_value
            == "SPN 1208 - Engine Oil Filter Intake Pressure (1208)"
        )
        assert (
            decoded[9].value
            == "Data Valid But Above Normal Operating Range - Moderately Severe Level"
        )
        assert decoded[9].raw == 16
        assert decoded[10].value == 0
        assert (
            decoded[11].value
            == "means convert SPNs per the Version 4 definition below"
        )

        # Second SPN
        assert decoded[12].raw == 0x3BBBB
        assert decoded[12].value == "SPN 244667"
        assert decoded[12].display_value == "SPN 244667 (244667)"
        assert decoded[12].value == "SPN 244667"
        assert decoded[13].raw == 0
        assert (
            decoded[13].value
            == "Data Valid But Above Normal Operational Range - Most Severe Level"
        )
        assert decoded[14].value == 0
        assert (
            decoded[15].value
            == "means convert SPNs per the Version 4 definition below"
        )

    def test_it_unpacks_PGN_64912_AdvertisedEngineTorqueCurve(self, pgn_64912):
        sut = pgn_64912
        assert sut.id == 64912
        assert sut.name == "Advertised Engine Torque Curve"

        # First byte is standard and "point count"
        # following points are (speed, torque) pairs of 16-bit numbers
        inp = b"\x20\x12\x34\x56\x78\x9a\xbc\xde\xf0"
        decoded = sut.decode(inp)

        assert decoded[0].value == "SAE J1995"
        assert decoded[1].value == 2
        assert decoded[2].value == 0x3412 * 0.125
        assert str(decoded[2]) == "SPN:3560: AETC Speed Value = 1666.25"
        assert decoded[2].display_value == "1666.25 rpm"
        assert decoded[3].value == 0x7856
        assert str(decoded[3]) == "SPN:3561: AETC Torque value = 30806"
        assert decoded[3].display_value == "30806 Nm"
        assert decoded[4].value == 0xBC9A * 0.125
        assert str(decoded[4]) == "SPN:3560: AETC Speed Value = 6035.25"
        assert decoded[5].value == 0xF0DE
        assert str(decoded[5]) == "SPN:3561: AETC Torque value = 61662"


def test_all_SPNs():
    try:
        with open("./decoda_spec.json") as f:
            spec = json.load(f)
    except FileNotFoundError:
        pytest.skip("Unable to find a spec file")

    missing_bit_length = []
    errors = []
    for idx, spn in enumerate(spec["SPNs"]):
        try:
            spn_from_dict(spn)
        except MissingBitLength as e:
            missing_bit_length.append(e)
        except Exception as e:
            errors.append((idx, spn, e))

    assert len(errors) == 0
    assert len(missing_bit_length) < 1200


def test_all_PGNs():
    try:
        with open("./decoda_spec.json") as f:
            spec = json.load(f)
    except FileNotFoundError:
        pytest.skip("Unable to find a spec file")
    spn_specs = spec["SPNs"]
    spn_specs = [
        spn for spn in spn_specs if spn.get("bit_length") not in ("", None)
    ]
    spns = Repo(SPN, spn_from_dict, spn_specs)

    errors = []
    count = 0
    for idx, pgn in enumerate(spec["PGNs"]):
        try:
            pgn_from_dict(pgn, spns)
            count += 1
        except Exception as e:
            errors.append((idx, pgn, e))

    for _, pgn, error in errors:
        print("{} : {}".format(error, pgn))
    assert len(errors) == 0


@pytest.mark.parametrize(
    "pgn_id",
    [
        59392,
        60416,
        64793,
        64888,
        64889,
        64920,
        64921,
        65168,
        65208,
        65209,
        65251,
    ],
)
def test_pgns_can_decode(pgn_id, spec):
    pgn = spec.PGNs.get_by_id(pgn_id)
    inp = bytes([1] * 200)  # it's really long
    pgn.decode(inp)


def test_it_can_load_from_a_spec_file():
    try:
        spec = load_from_file("./decoda_spec.json")
    except FileNotFoundError:
        pytest.skip("Unable to find a spec file")
    assert spec.Manufacturers.get_by_id(2).name == "Allison Transmission, Inc."
    assert (
        spec.SPNs.get_by_id(22).name
        == "Engine Extended Crankcase Blow-by Pressure"
    )
    assert spec.PGNs.get_by_id(0).name == "Torque/Speed Control 1"



import argparse
import json
import re
import sys

from decoda.sae_spec_converter.create_j1939db_json import J1939daConverter


def update_resolution(d):
    d = d.copy()

    resolution = d.get("resolution", "").lower()
    units = d.get("units", "").lower()
    if not resolution:
        return d
    if resolution in {"variant determined", "ascii"}:
        return d
    if units in {"ascii", "binary", "bit", "bit-mapped"}:
        return d

    # Try using "per" for splitting, this appears to be the new way
    parts = resolution.split(f"{units} per ")
    try:
        d["resolution"] = eval(parts[0].replace(" ", ""))
        return d
    except Exception:
        if units == "" and "/" not in resolution:
            return d
        pass

    # Fall-back to old mappings
    # map known mismatches
    # NB: units are lower case from above!
    units = {
        "ratio": "/",
        "weeks": "week",  # SPN 915
        "months": "month",  # SPN 963
        "years": "year",  # SPN 964
        "": "/",  # SPN 1255
        "sa": "source address",  # SPN 1480
        "turns": "turn",  # SPN 1811
        "ohm/v": "ohm per",  # SPN 8431
    }.get(units, units)

    if units not in resolution:
        raise ValueError(f"missing units in resolution: {d}")

    parts = resolution.split(units)
    try:
        d["resolution"] = eval(parts[0].replace(" ", ""))
    except Exception:
        print(d)
        raise

    return d


# Remove all characters that are not likely to be part of a number
def _strip_not_likely_numericals(value):
    # Strip hyphens not followed by digit, like in "Watt-hour"
    value = re.sub(r"-(?!\d)", "", value)
    # Strip everything not likely to be part of a number
    return re.sub(r"[^\d\.\-]", "", value)


def update_offset(d):
    d = d.copy()
    if d["offset"] not in {
        "",
        "Variant",
        "Variant Determined",
        "Data Specific",
    }:
        try:
            offset = _strip_not_likely_numericals(d["offset"])
            d["offset"] = eval(offset.split()[0])
        except:
            print(d)
            raise
    return d


def update_datarange(d):
    d = d.copy()
    try:
        if d["units"] == "ASCII":
            pass
        elif " to " in d["data_range"]:
            data_range = d["data_range"]
            if d["units"]:
                data_range = data_range.split(d["units"])[0].strip()
            low, high = data_range.split(" to ")
            # Remove all characters that are not likely to be part of a number
            low = _strip_not_likely_numericals(low)
            high = _strip_not_likely_numericals(high)
            d["data_range"] = {
                "min": float(low.replace(",", "").replace(" ", "")),
                "max": float(high.replace(",", "").replace(" ", "")),
            }
    except:
        print(d)
        raise

    return d


#      "length": "Variable - up to 100 bytes",
#      "length": "Variable - up to 1728 bytes followed by an \"*\" delimiter",
#      "length": "Variable - up to 200 bytes followed by an NULL delimiter",
#      "length": "Variable - up to 200 bytes followed by an \"*\" delimiter",
#      "length": "Variable - up to 200 characters",
#      "length": "Variable - up to 25 bytes followed by an \"*\" delimiter",
#      "length": "Variable - up to 32 bytes followed by an \"*\" delimiter",
#      "length": "Variable - up to 5 bytes followed by an \"*\" delimiter",
#      "length": "Variable",
def append_bitlength(d):
    d = d.copy()
    length = d["length"]
    if length == "":
        return d

    parts = length.split(" ")

    if "Variable" in length:
        d["variable"] = True
        d["bit_length"] = int(parts[4]) * 8
        if parts[-1] == "delimiter":
            d["delimiter"] = hex(ord("*")) if "*" in parts[-2] else hex(0)
        length_spn = re.search(
            "The length.*must be reported using SPN (\\d+) ", d["description"]
        )
        if length_spn:
            d["length_spn"] = int(length_spn.group(1))

        return d

    if parts[1] in {"byte", "bytes"}:
        d["bit_length"] = int(parts[0]) * 8
    elif parts[1] in {"bit", "bits"}:
        d["bit_length"] = int(parts[0])
    else:
        raise ValueError(f"Unhandled length for SPN: {d}")
    return d


def extract_encodings(d):
    # We really only want the first encounter of an encoding, as that is the
    # shortest. This works around what I consider to be a bug in
    # J1939daConverter.
    class OnlyTakeFirst:
        encodings = {}
        extras = {}

        def update(self, d):
            for key, val in d.items():
                update_dict = (
                    self.encodings
                    if key not in self.encodings
                    else self.extras
                )
                update_dict[key] = val

    take_first = OnlyTakeFirst()
    try:
        J1939daConverter.create_bit_object_from_description(
            d["description"], take_first
        )
    except Exception:
        print(d)
        raise

    if not take_first.encodings:
        return d

    d = d.copy()

    description = d["description"]
    # Apply the same replacement treatments
    description = re.sub(r"[ ]+", " ", description)
    description = re.sub(r"[ ]?\-\-[ ]?", " = ", description)
    description = re.sub(
        r"^(\d+)\s*-\s*(?!\=)", r"\1 = ", description
    )  # replace hyphen with = when in form "123 - "
    description_lower = description.lower()
    encodings = {}
    for key, value in take_first.encodings.items():
        # The create_bit_object_from_description uses a .lower() in part of its
        # cleaning of the strings. I don't like that, so do some tricks to get
        # the properly capitalised strings back
        start_idx = description_lower.index(value.lower())
        encodings[key] = description[start_idx : start_idx + len(value)]

    d["encodings"] = encodings

    return d


def enrich_spns(spns):
    result = []
    for spn in spns:
        spn = update_resolution(spn)
        spn = update_offset(spn)
        spn = update_datarange(spn)
        spn = append_bitlength(spn)
        spn = extract_encodings(spn)
        result.append(spn)
    return result


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "input_file",
        nargs="?",
        type=argparse.FileType("r"),
        default=sys.stdout,
        help="the json file to use as input",
    )
    parser.add_argument(
        "output_file",
        nargs="?",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="the json file to use as output",
    )
    parser.add_argument("--pretty", default=False, action="store_true")
    args = parser.parse_args()

    spec = json.load(args.input_file)
    spec["SPNs"] = enrich_spns(spec["SPNs"])
    json.dump(spec, args.output_file, indent=2 if args.pretty else None)


if __name__ == "__main__":
    main()



import pytest

from decoda import *


@pytest.mark.parametrize(
    ["can_id", "priority", "pgn", "sa", "da"],
    [
        (0x0CF004FE, 3, 61444, Address(254), BroadcastAddress),
        (0x98FEF4FE, 6, 65268, Address(254), BroadcastAddress),
        (0x18F00930, 6, 61449, Address(48), BroadcastAddress),
        (0x18F00131, 6, 61441, Address(49), BroadcastAddress),
        (0x00DEAA50, 0, 0xDE00, Address(0x50), Address(0xAA)),
        (0x00000102, 0, 0, Address(2), Address(1)),
    ],
)
def test_it_unpacks_can_id(can_id, priority, pgn, sa, da):
    assert (priority, pgn, sa, da) == parts_from_can_id(can_id), hex(pgn)


@pytest.mark.parametrize(
    ["pf", "ps", "da"],
    [
        (0x01, 0x00, Address(0)),
        (0x01, 0x01, Address(1)),
        (0xEF, 0x09, Address(9)),
        (0xF0, 0x00, BroadcastAddress),
        (0xF1, 0x00, BroadcastAddress),
        (0xF2, 0x00, BroadcastAddress),
    ],
)
def test_it_can_get_destination_address_from_pf_and_ps(pf, ps, da):
    assert Address.from_pf_and_ps(pf, ps) == da



from unittest.mock import patch, sentinel

import pytest

from decoda import *

BL8 = BitLength(8)


class TestScalarValue:
    def test_it_decodes_without_anything(self):
        sv = scalar_value_from_dict({"bit_length": 8})
        result, _ = sv.decode_from_raw(b"\0", "1", False, -1)
        assert result == {"value": 0, "raw": 0, "display_value": "0"}
        result, _ = sv.decode_from_raw(b"\x7b", "1", False, -1)
        assert result == {"value": 123, "raw": 123, "display_value": "123"}

    def test_it_includes_units_in_display(self):
        sv = scalar_value_from_dict({"units": "XYZs", "bit_length": 8})
        result, _ = sv.decode_from_raw(b"\x7b", "1", False, -1)
        assert result["display_value"] == "123 XYZs"

    def test_it_scales_if_included(self):
        sv = scalar_value_from_dict({"resolution": 5, "bit_length": 8})
        result, _ = sv.decode_from_raw(b"\x0a", "1", False, -1)
        assert result["raw"] == 10
        assert result["value"] == 50

    def test_it_scales_and_offsets(self):
        sv = scalar_value_from_dict(
            {
                "resolution": 0.125,
                "offset": -123,
                "bit_length": 16,
            }
        )  # yapf: disable
        result, _ = sv.decode_from_raw(b"\x0a\0", "1", False, -1)
        assert result["value"] == (-123 + 10 * 0.125)
        result, _ = sv.decode_from_raw(b"\x64\0", "1", False, -1)
        assert result["value"] == (-123 + 100 * 0.125)
        result, _ = sv.decode_from_raw(b"\xe7\x03", "1", False, -1)
        assert result["value"] == (-123 + 999 * 0.125)

    def test_it_clips_to_min_after_scaling_and_offset(self):
        sv = scalar_value_from_dict(
            {
                "units": "XYZs",
                "data_range": {"min": 10},
                "resolution": 2,
                "offset": 5,
                "bit_length": 8,
            }
        )  # yapf: disable
        result, _ = sv.decode_from_raw(b"\2", "1", False, -1)
        assert result == {
            "value": 10,
            "raw": 2,
            "display_value": "10 XYZs (encoded value 9 XYZs was clipped to min)",
        }

    def test_it_clips_to_max_after_scaling_and_offset(self):
        sv = scalar_value_from_dict(
            {
                "units": "XYZs",
                "data_range": {"max": 10},
                "resolution": 3.3,
                "offset": 5,
                "bit_length": 8,
            }
        )  # yapf: disable
        result, _ = sv.decode_from_raw(b"\2", "1", False, -1)
        assert result == {
            "value": 10,
            "raw": 2,
            "display_value": "10 XYZs (encoded value 11.6 XYZs was clipped to max)",
        }  # yapf: disable

    def test_it_raises_error_if_min_and_max_not_correct(self):
        with pytest.raises(ValueError) as e:
            scalar_value_from_dict(
                {"data_range": {"min": 10, "max": 10}, "bit_length": 8}
            )
        assert str(e.value) == "min must be less than max"

    def test_it_ignores_min_and_max_if_requested(self):
        sv = scalar_value_from_dict(
            {
                "units": "XYZs",
                "data_range": {"min": 1, "max": 10},
                "resolution": 1,
                "offset": -5,
                "bit_length": 8,
            }
        )  # yapf: disable
        result, _ = sv.decode_from_raw(
            b"\0", "1", False, -1, ignore_range=True
        )
        assert result["value"] == -5
        result, _ = sv.decode_from_raw(
            b"\x14", "1", False, -1, ignore_range=True
        )
        assert result["value"] == 15

    @pytest.mark.parametrize(
        ["bit_length", "byte_stream", "display_value_suffix"],
        [
            (8, b"\xff", ""),
            (16, b"\x00\xff", " (0)"),
            (16, b"\x05\xff", " (5)"),
            (32, b"\x00\x00\x00\xff", " (0)"),
            (32, b"\x05\x00\x00\xff", " (5)"),
        ],
    )
    def test_it_reports_not_available_when_in_FF_range(
        self, bit_length, byte_stream, display_value_suffix
    ):
        sut = scalar_value_from_dict(
            {
                "units": "XYZs",
                "bit_length": bit_length,
            }
        )
        result, _ = sut.decode_from_raw(
            byte_stream, "1", False, -1, ignore_range=True
        )
        assert result["value"] == "Not available"
        assert (
            result["display_value"] == "Not available" + display_value_suffix
        )

    @pytest.mark.parametrize(
        ["bit_length", "byte_stream", "display_value_suffix"],
        [
            (8, b"\xfe", ""),
            (16, b"\x00\xfe", " (0)"),
            (16, b"\x05\xfe", " (5)"),
            (32, b"\x00\x00\x00\xfe", " (0)"),
            (32, b"\x05\x00\x00\xfe", " (5)"),
        ],
    )
    def test_it_reports_error_indicator_when_in_FE_range(
        self, bit_length, byte_stream, display_value_suffix
    ):
        sut = scalar_value_from_dict(
            {
                "units": "XYZs",
                "bit_length": bit_length,
            }
        )
        result, _ = sut.decode_from_raw(
            byte_stream, "1", False, -1, ignore_range=True
        )
        assert result["value"] == "Error indicator"
        assert (
            result["display_value"] == "Error indicator" + display_value_suffix
        )

    @pytest.mark.parametrize(
        ["bit_length", "byte_stream", "display_value_suffix"],
        [
            (8, b"\xfb", ""),
            (16, b"\x00\xfb", " (0)"),
            (16, b"\x05\xfb", " (5)"),
            (16, b"\xff\xfd", " (767)"),
            (32, b"\x00\x00\x00\xfb", " (0)"),
            (32, b"\x05\x00\x00\xfb", " (5)"),
            (32, b"\xff\xff\xff\xfd", " (50331647)"),
        ],
    )
    def test_it_reports_parameter_indicator_when_in_FB_to_FE_range(
        self, bit_length, byte_stream, display_value_suffix
    ):
        sut = scalar_value_from_dict(
            {
                "units": "XYZs",
                "bit_length": bit_length,
            }
        )
        result, _ = sut.decode_from_raw(
            byte_stream, "1", False, -1, ignore_range=True
        )
        assert result["value"] == "Parameter specific indicator"
        assert (
            result["display_value"]
            == "Parameter specific indicator" + display_value_suffix
        )

    def test_it_requires_positive_max_len(self):
        with pytest.raises(ValueError) as e:
            scalar_value_from_dict({"bit_length": -1})
        assert str(e.value) == "must have positive bit length"


class TestEncodedValue:
    def test_it_decodes_to_available_encodings(self):
        ev = EncodedValue({0: "Off", 1: "On"}, BitLength(1))
        result, _ = ev.decode_from_raw(b"\0", "1", False, -1)
        assert result == {"value": "Off", "raw": 0, "display_value": "Off (0)"}
        result, _ = ev.decode_from_raw(b"\1", "1", False, -1)
        assert result == {"value": "On", "raw": 1, "display_value": "On (1)"}

    def test_it_decodes_unknown_encodings_to_something(self):
        ev = EncodedValue({0: "Off", 1: "On"}, BitLength(2))
        result, _ = ev.decode_from_raw(b"\2", "1", False, -1)
        assert result == {
            "value": None,
            "raw": 2,
            "display_value": "No encoding (2)",
        }

    def test_it_raises_error_if_there_are_no_encodings(self):
        with pytest.raises(ValueError) as e:
            EncodedValue({}, None)
        assert str(e.value) == "must have at least 1 encoding"

    def test_it_handles_ints_and_strings_in_encoding_keys(self):
        ev = EncodedValue({"0": "Off", 1: "On"}, BitLength(1))
        result, _ = ev.decode_from_raw(b"\0", "1", False, -1)
        assert result == {"value": "Off", "raw": 0, "display_value": "Off (0)"}
        result, _ = ev.decode_from_raw(b"\1", "1", False, -1)
        assert result == {"value": "On", "raw": 1, "display_value": "On (1)"}

    def test_it_handles_ranges_in_encoding_keys(self):
        ev = EncodedValue(
            {
                "0": "North",
                "1": "South",
                "2": "East",
                "3": "West",
                "4": "In",
                "5": "Out",
                "6-15": "Agency defined",
            },
            BitLength(4),
        )

        result, _ = ev.decode_from_raw(b"\0", "1", False, -1)
        assert result == {
            "value": "North",
            "raw": 0,
            "display_value": "North (0)",
        }
        result, _ = ev.decode_from_raw(b"\x0a", "1", False, -1)
        assert result == {
            "value": "Agency defined",
            "raw": 10,
            "display_value": "Agency defined (10)",
        }
        result, _ = ev.decode_from_raw(b"\x0e", "1", False, -1)
        assert result == {
            "value": "Agency defined",
            "raw": 14,
            "display_value": "Agency defined (14)",
        }

    def test_it_requires_positive_max_len(self):
        encodings = {"0": "Don't care"}
        with pytest.raises(ValueError) as e:
            EncodedValue(encodings, BitLength(None))
        assert str(e.value) == "must have positive bit length"

        with pytest.raises(ValueError) as e:
            EncodedValue(encodings, BitLength(-1))
        assert str(e.value) == "must have positive bit length"


class TestBitLength:
    @pytest.mark.parametrize("bit_length", [1, 7, 9])
    def test_it_cannot_be_constructed_with_non_byte_boundary_length(
        self, bit_length
    ):
        with pytest.raises(ValueError) as e:
            BitLength(bit_length, True)
        assert str(
            e.value
        ) == "bit_length ({}) must be an 8-bit multiple".format(bit_length)


class TestTextValue:
    def test_it_decodes(self):
        tv = TextValue(BitLength(None, True), "*")
        result, _ = tv.decode_from_raw(bytes("hello", "ascii"), "1", False, -1)
        assert result == {
            "value": "hello",
            "raw": "hello",
            "display_value": "hello",
        }

    def test_it_raises_error_if_not_variable_and_insufficient_bytes(self):
        tv = TextValue(BitLength(5 * 8))

        with pytest.raises(ValueError) as e:
            tv.decode_from_raw(b"\1", "1", False, -1)

        assert str(e.value) == "insufficent bytes"

    def test_it_raises_error_if_variable_and_neither_delimiter_or_length_spn_supplied(
        self,
    ):
        with pytest.raises(ValueError) as e:
            TextValue(BitLength(None, True))

        assert (
            str(e.value)
            == "Variable length SPN must specify delimiter or length SPN"
        )

    @pytest.mark.parametrize("delimiter", ["0x00", "0x2a", "0xa"])
    def test_it_uses_provided_delimiter(self, delimiter):
        tv = TextValue(BitLength(None, True), delimiter)
        inp = (
            bytes("hello", "ascii")
            + bytes([int(delimiter, 16)])
            + bytes("world", "ascii")
        )
        result, end_byte_idx = tv.decode_from_raw(inp, "1", False, -1)
        assert result == {
            "value": "hello",
            "raw": "hello",
            "display_value": "hello",
        }
        assert end_byte_idx == 5

    @pytest.mark.parametrize(
        ["length", "expected", "expected_idx"],
        [
            (4, "hell", 3),
            (5, "hello", 4),
        ],
    )
    def test_it_decodes_using_already_decoded_length(
        self, length, expected, expected_idx
    ):
        tv = TextValue(BitLength(None, True), None, 1234)
        mockOneTwoThreeFourResult = DecodedSPN.build(
            spn=SPN(id=1234, name=None, description=None, value_decoder=None),
            value=length,
            raw=None,
            display_value=None,
        )
        result, end_byte_idx = tv.decode_from_raw(
            bytes("helloworld", "ascii"),
            "1",
            False,
            -1,
            already_decoded=[mockOneTwoThreeFourResult],
        )
        assert result["value"] == expected
        assert end_byte_idx == expected_idx


class TestByteArrayValue:
    def test_it_decode_fixed_bit_length(self):
        # SPN 2550 - Manufacturer Specific Information (PropA_PDU1)
        # Bitlength - 14280
        bv = ByteArrayValue(BitLength(14280))
        val = b"\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a"
        result, _ = bv.decode_from_raw(val, "1", False, -1)
        assert result["value"] == val
        assert result["raw"] == val
        assert result["display_value"] == "0102030405060708090a"

    def test_it_decode_variable_length(self):
        bv = ByteArrayValue(BitLength(14280, True))
        val = b"\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a"
        result, _ = bv.decode_from_raw(val, "1", False, -1)
        assert result["value"] == val
        assert result["raw"] == val
        assert result["display_value"] == "0102030405060708090a"


class TestSPNDecoding:
    def test_it_decodes_scalar_values(self):
        spn = SPN(1, "Name", "Desc", scalar_value_from_dict({"bit_length": 8}))
        result, _ = spn.decode(bytes([0]), "1", False, -1)
        assert result == DecodedSPN.build(spn, 0, 0, "0")
        result, _ = spn.decode(bytes([123]), "1", False, -1)
        assert result == DecodedSPN.build(spn, 123, 123, "123")

    def test_it_decodes_encoded_values(self):
        spn = SPN(1, "Name", "Desc", EncodedValue({1: "On", 0: "Off"}, BL8))

        result, _ = spn.decode(bytes([0]), "1", False, -1)
        assert result == DecodedSPN.build(spn, 0, "Off", "Off (0)")
        result, _ = spn.decode(bytes([1]), "1", False, -1)
        assert result == DecodedSPN.build(spn, 1, "On", "On (1)")

    def test_it_decodes_text_values(self):
        bitLen40 = BitLength(40)
        spn = SPN(1, "Name", "Desc", TextValue(bitLen40))

        result, _ = spn.decode(bytes("hello", "ascii"), "1", False, -1)
        assert result == DecodedSPN.build(spn, "hello", "hello", "hello")

    def test_it_raises_an_error_if_value_does_not_have_enough_bits(self):
        spn = SPN(
            1, "Name", "Desc", scalar_value_from_dict({"bit_length": 16})
        )

        with pytest.raises(ValueError) as e:
            spn.decode(bytes([0]), "1", False, -1)

        assert (
            str(e.value)
            == "not enough bits for SPN - need 16, from 0.0 in payload[b'00']"
        )


class TestSPNConstruction:
    @patch("decoda.spec_loader.scalar_value_from_dict")
    def test_it_builds_a_scalar_decoding_spn(self, scalar_value_from_dict):
        d = {
            "id": 4151,
            "name": "A name of an SPN",
            "bit_length": 16,
            "data_range": {"min": -273, "max": 1734.96875},
            "offset": -273,
            "resolution": 0.03125,
            "units": "deg C",
        }  # yapf: disable
        result = spn_from_dict(d)
        assert result.id == 4151
        assert result.name == "A name of an SPN"
        assert result.description == ""
        assert result.value_decoder == scalar_value_from_dict.return_value
        scalar_value_from_dict.assert_called_with(d)

    @patch("decoda.spec_loader.EncodedValue")
    def test_it_builds_an_encoding_decoding_spn(self, EncodedValue):
        spn = spn_from_dict(
            {
                "id": 4309,
                "name": "A name of an SPN",
                "description": "",
                "bit_length": 2,
                "offset": "",
                "encodings": {
                    "0": "Reverse",
                    "1": "Forward",
                    "2": "Error indication",
                    "3": "Not available",
                },
            }
        )
        assert spn.id == 4309
        assert spn.name == "A name of an SPN"
        assert spn.description == ""
        assert spn.value_decoder == EncodedValue.return_value
        EncodedValue.assert_called_with(
            {
                "0": "Reverse",
                "1": "Forward",
                "2": "Error indication",
                "3": "Not available",
            },
            BitLength(2),
        )

    @patch("decoda.spec_loader.TextValue")
    def test_it_builds_a_text_decoding_spn(self, TextValue):
        spn = spn_from_dict(
            {
                "id": 4254,
                "name": "A name of an SPN",
                "description": "",
                "bit_length": 200,
                "offset": "",
                "data_range": {"min": 0, "max": 255},
                "units": "ASCII",  # THIS IS IMPORTANT
            }
        )  # yapf: disable
        assert spn.id == 4254
        assert spn.name == "A name of an SPN"
        assert spn.description == ""
        assert spn.value_decoder == TextValue.return_value
        TextValue.assert_called_with(
            BitLength(200), delimiter=None, length_spn=None
        )

    @patch("decoda.spec_loader.TextValue")
    def test_it_builds_a_text_decoding_spn_with_variable_length(
        self, TextValue
    ):
        spn = spn_from_dict(
            {
                "id": 4254,
                "name": "A name of an SPN",
                "bit_length": 200,
                "units": "ASCII",  # THIS IS IMPORTANT
                "variable": True,
                "delimiter": "*",
            }
        )  # yapf: disable
        assert spn.value_decoder == TextValue.return_value
        TextValue.assert_called_with(
            BitLength(200, True), delimiter="*", length_spn=None
        )

    @patch("decoda.spec_loader.TextValue")
    def test_it_builds_a_text_decoding_spn_with_length_spn(self, TextValue):
        spn = spn_from_dict(
            {
                "id": 4254,
                "name": "A name of an SPN",
                "bit_length": 200,
                "units": "ASCII",  # THIS IS IMPORTANT
                "variable": True,
                "length_spn": 1234,
            }
        )  # yapf: disable
        assert spn.value_decoder == TextValue.return_value
        TextValue.assert_called_with(
            BitLength(200, True), delimiter=None, length_spn=1234
        )



import argparse
import datetime
import json
import sys

import defusedxml
import xlrd2 as xlrd
from defusedxml.common import EntitiesForbidden
import pandas as pd


class XlrdAdapter:
    def __init__(self, wb):
        self._wb = wb

    def find_first_sheet_by_name(self, sheet_names):
        if not isinstance(sheet_names, list):
            sheet_names = [sheet_names]
        for sheet_name in sheet_names:
            if sheet_name in self._wb.sheet_names():
                sheet = self._wb.sheet_by_name(sheet_name)
                return sheet
        return None

    def get_header_row(self, sheet):
        row_num = 3 if sheet.row_values(0)[3].strip() == "" else 0

        header_row = sheet.row_values(row_num)
        normalised_headers = [
            header.upper().replace(" ", "_") for header in header_row
        ]
        return normalised_headers, row_num

    def convert_to_datetime(self, value):
        return datetime.datetime(
            *xlrd.xldate_as_tuple(value, self._wb.datemode)
        )

    def number_rows(self, sheet):
        return sheet.nrows

    def row_values(self, sheet, row_index):
        return sheet.row_values(row_index)


class PandasAdapter:
    def __init__(self, filename):
        self._filename = filename

    @staticmethod
    def create(filename, **kwargs):
        return PandasAdapter(filename)

    def find_first_sheet_by_name(self, sheet_names):
        if not isinstance(sheet_names, list):
            sheet_names = [sheet_names]
        for sheet_name in sheet_names:
            try:
                df = pd.read_excel(
                    self._filename,
                    sheet_name=sheet_name,
                    index_col=None,
                    header=None,
                )
                df = df.where(pd.notnull(df), "")
                # Replace occurrences of _x000D_ with an empty string (i.e. pandas' version of \r
                df = df.replace("_x000D_", "", regex=True)
                return df
            except ImportError:
                raise
            except Exception:
                pass
        return None

    def get_header_row(self, sheet):
        indicator_cell_value = sheet.values[0, 3]
        row_num = 0 if indicator_cell_value else 3

        header_row = sheet.values[row_num]
        normalised_headers = [
            (header or "").upper().replace(" ", "_") for header in header_row
        ]
        return normalised_headers, row_num

    def convert_to_datetime(self, value):
        return value

    def number_rows(self, sheet):
        return len(sheet)

    def row_values(self, sheet, row_index):
        return sheet.iloc[row_index]


def secure_open_workbook(filename, **kwargs):
    if filename.lower().endswith("xlsx"):
        return PandasAdapter.create(filename, read_only=True)

    defusedxml.defuse_stdlib()
    try:
        return XlrdAdapter(xlrd.open_workbook(filename=filename, **kwargs))
    except EntitiesForbidden:
        raise ValueError("Please use an excel file without XEE")


def get_header_index_any_match(headers, names):
    if not isinstance(names, list):
        names = [names]
    for n in names:
        try:
            return headers.index(n)
        except ValueError:
            continue
    raise ValueError(f"Unabled to find column named one of: {names}")


def dedup_and_flag_discrepancies(items, key="id"):
    by_key = {}
    identical_dups = 0
    for item in items:
        k = item[key]
        if k in by_key:
            if by_key[k] != item:
                print(f"Warning, mismatch found for {item}")
            else:
                identical_dups += 1
        else:
            by_key[k] = item
    print(f"Removed {identical_dups} identical duplicates")
    return [by_key[k] for k in sorted(by_key.keys())]


def extract_manfacturers(wb):
    sheet = wb.find_first_sheet_by_name(["Manufacturer IDs (B10)"])

    headers, row_num = wb.get_header_row(sheet)
    id_col = get_header_index_any_match(headers, "MFR_ID")
    name_col = get_header_index_any_match(headers, "MANUFACTURER")
    location_col = get_header_index_any_match(headers, "LOCATION")
    last_modified_col = get_header_index_any_match(
        headers, "DATE_CREATED_OR_LAST_MODIFIED"
    )

    result = []
    for i in range(row_num + 1, wb.number_rows(sheet)):
        row = wb.row_values(sheet, i)
        last_modified = row[last_modified_col]

        if not isinstance(last_modified, str):
            last_modified = wb.convert_to_datetime(last_modified)

        if last_modified:
            last_modified = last_modified.date().isoformat()

        result.append(
            {
                "id": int(row[id_col]),
                "name": str(row[name_col]),
                "location": str(row[location_col]),
                "created_or_last_modified": last_modified,
            }
        )
    return result


def extract_spns(wb):
    sheet = wb.find_first_sheet_by_name(["SPNs & PGNs", "SPs & PGs"])
    headers, row_num = wb.get_header_row(sheet)

    id_col = get_header_index_any_match(headers, "SPN")
    name_col = get_header_index_any_match(headers, ["SPN_NAME", "SP_LABEL"])
    description_col = get_header_index_any_match(
        headers, ["SPN_DESCRIPTION", "SP_DESCRIPTION"]
    )
    length_col = get_header_index_any_match(
        headers, ["SPN_LENGTH", "SP_LENGTH"]
    )
    resolution_col = get_header_index_any_match(
        headers, ["RESOLUTION", "SCALING"]
    )
    offset_col = get_header_index_any_match(headers, "OFFSET")
    data_range_col = get_header_index_any_match(headers, "DATA_RANGE")
    operational_range_col = get_header_index_any_match(
        headers, "OPERATIONAL_RANGE"
    )
    units_col = get_header_index_any_match(headers, ["UNITS", "UNIT"])
    document_col = get_header_index_any_match(
        headers, ["SPN_DOCUMENT", "SP_DOCUMENT"]
    )

    result = []
    for i in range(row_num + 1, wb.number_rows(sheet)):
        row = wb.row_values(sheet, i)
        spn_id = row[id_col]
        if spn_id == "" or spn_id is None:
            continue

        result.append(
            {
                "id": int(spn_id),
                "name": str(row[name_col]),
                "description": str(row[description_col]),
                "length": str(row[length_col]),
                "resolution": str(row[resolution_col]),
                "offset": str(row[offset_col]),
                "data_range": str(row[data_range_col]),
                "operational_range": str(row[operational_range_col]),
                "units": str(row[units_col]),
                "source_document": str(row[document_col]),
            }
        )

    return dedup_and_flag_discrepancies(result)


def int_or_str(s):
    try:
        return int(s)
    except:
        return s


def extract_pgns(wb):
    sheet = wb.find_first_sheet_by_name(["SPNs & PGNs", "SPs & PGs"])
    headers, row_num = wb.get_header_row(sheet)

    id_col = get_header_index_any_match(headers, "PGN")
    name_col = get_header_index_any_match(
        headers, ["PARAMETER_GROUP_LABEL", "PG_LABEL"]
    )
    acronym_col = get_header_index_any_match(
        headers, ["ACRONYM", "PG_ACRONYM"]
    )
    description_col = get_header_index_any_match(
        headers, ["PGN_DESCRIPTION", "PG_DESCRIPTION"]
    )
    length_col = get_header_index_any_match(
        headers,
        ["PGN_DATA_LENGTH", "PG_DATA_LENGTH", "PG_DATA_MINIMUM_LENGTH"],
    )
    rate_col = get_header_index_any_match(headers, "TRANSMISSION_RATE")

    spn_id_col = get_header_index_any_match(headers, "SPN")
    spn_position_col = get_header_index_any_match(
        headers, ["SPN_POSITION_IN_PGN", "SP_POSITION_IN_PG"]
    )
    document_col = get_header_index_any_match(
        headers, ["PGN_DOCUMENT", "PG_DOCUMENT"]
    )

    result = []
    current_pgn = {"id": "bogus"}
    for i in range(row_num + 1, wb.number_rows(sheet)):
        row = wb.row_values(sheet, i)
        pgn_id = row[id_col]
        if pgn_id == "" or pgn_id is None:
            continue

        pgn_id = int(pgn_id)
        if pgn_id != current_pgn["id"]:
            # new record found
            current_pgn = {
                "id": pgn_id,
                "name": str(row[name_col]),
                "acronym": str(row[acronym_col]),
                "description": str(row[description_col]),
                "length": int_or_str(row[length_col]),
                "rate": str(row[rate_col]),
                "source_document": str(row[document_col]),
                "spns": [],
            }
            result.append(current_pgn)

        # Only append SPNs that are valid
        spn_id = row[spn_id_col] or ""
        if spn_id != "":
            # TODO - check for consistency?
            current_pgn["spns"].append(
                {"id": int(spn_id), "start_pos": str(row[spn_position_col])}
            )

    return dedup_and_flag_discrepancies(result)


def extract_industry_groups(wb):
    sheet = wb.find_first_sheet_by_name(["Industry Groups (B1)"])
    headers, row_num = wb.get_header_row(sheet)

    id_col = get_header_index_any_match(headers, "INDUSTRY_GROUP_ID")
    name_col = get_header_index_any_match(
        headers, "INDUSTRY_GROUP_DESCRIPTION"
    )

    result = []
    for i in range(row_num + 1, wb.number_rows(sheet)):
        row = wb.row_values(sheet, i)
        result.append(
            {
                "id": int(row[id_col]),
                "name": str(row[name_col]),
            }
        )
    return result


def extract_source_addresses(wb):
    result = {}

    for sa_type, sheet_names in [
        ("Global", ["Global Source Addresses (B2)"]),
        ("IG1", ["IG1 Source Addresses (B3)"]),
        ("IG2", ["IG2 Source Addresses (B4)"]),
        ("IG3", ["IG3 Source Addresses (B5)"]),
        ("IG4", ["IG4 Source Addresses (B6)"]),
        ("IG5", ["IG5 Source Addresses (B7)"]),
    ]:

        sheet = wb.find_first_sheet_by_name(sheet_names)
        headers, row_num = wb.get_header_row(sheet)

        id_col = get_header_index_any_match(headers, "SOURCE_ADDRESS_ID")
        name_col = get_header_index_any_match(headers, ["NAME", "FUNCTION"])

        addresses = {}
        for i in range(row_num + 1, wb.number_rows(sheet)):
            row = wb.row_values(sheet, i)
            id_val = int(row[id_col])
            name_val = str(row[name_col])

            if name_val.startswith("thru "):
                parts = name_val.split(" ")
                until = int(parts[1]) + 1
                name_val = " ".join(["Reserved"] + parts[4:])
            else:
                until = id_val + 1

            for id in range(id_val, until):
                addresses[id] = name_val

        result[sa_type] = addresses
    return result


def spec_from_workbook(wb: xlrd.Book):
    return {
        "Manufacturers": extract_manfacturers(wb),
        "PGNs": extract_pgns(wb),
        "SPNs": extract_spns(wb),
        "IndustryGroups": extract_industry_groups(wb),
        "SourceAddresses": extract_source_addresses(wb),
    }


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "digital_annex_xls",
        type=str,
        help="the J1939 Digital Annex .xls excel file used as input",
    )
    parser.add_argument(
        "json_file",
        nargs="?",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="the json file to use as output",
    )
    parser.add_argument("--pretty", default=False, action="store_true")
    args = parser.parse_args()

    print(args.digital_annex_xls)
    defusedxml.defuse_stdlib()
    wb = secure_open_workbook(filename=args.digital_annex_xls, on_demand=True)
    spec = spec_from_workbook(wb)
    json.dump(spec, args.json_file, indent=2 if args.pretty else None)


if __name__ == "__main__":
    main()



# Copyright Andrew Dodd
from numbers import Number


class UnknownReferenceError(Exception):
    pass


class CouldBeAnyLength(Exception):
    pass


class UnsupportedLocationSpec(ValueError):
    pass


class MissingBitLength(ValueError):
    pass


class ScalarRangeError(ValueError):
    def __init__(self, message: str, error_code: Number = None):
        super().__init__(message)
        self.error_code = error_code

    def value(self):
        return str(self)

    def display_value(self):
        if self.error_code is None:
            return self.value()
        return f"{self.value()} ({self.error_code})"


class NotAvaiableRangeError(ScalarRangeError):
    def __init__(self, error_code: Number):
        super().__init__("Not available", error_code)


class ErrorIndicatorRangeError(ScalarRangeError):
    def __init__(self, error_code: Number):
        super().__init__("Error indicator", error_code)


class ParameterSpecificIndicatorError(ScalarRangeError):
    def __init__(self, error_code: Number):
        super().__init__("Parameter specific indicator", error_code)



import pytest

from decoda.sae_spec_converter.enrich_spec import (
    update_datarange,
    update_offset,
)


@pytest.mark.parametrize(
    ["units", "data_range", "expected"],
    [
        ("", "", ""),
        ("abcs", "12 to 23 abcs", {"min": 12, "max": 23}),
        ("abcs", "1.0 to 9.0 abcs", {"min": 1.0, "max": 9.0}),
        ("abcs", "1 234 to 9 999 abcs", {"min": 1234, "max": 9999}),
        (
            "don't care",
            "0 to 125.498 046 875 km/L",
            {"min": 0, "max": 125.498046875},
        ),
    ],
)
def test_update_datarange_converts_as_expected(units, data_range, expected):
    assert (
        update_datarange({"units": units, "data_range": data_range})[
            "data_range"
        ]
        == expected
    )


@pytest.mark.parametrize(
    ["offset", "expected"],
    [
        ("", ""),
        ("-1234", -1234),
        ("-1 234", -1234),
        ("-1,234 mm", -1234),
        ("0 Watt-hour", 0),
    ],
)
def test_it_updates_offet(offset, expected):
    assert update_offset({"offset": offset})["offset"] == expected



import pytest

from decoda import *

BL8 = BitLength(8)


def test_it_decodes_single_byte_spns(pgn_1792: PGN):
    results = pgn_1792.decode((0x0).to_bytes(8, "big"))
    result_4087 = results[1]
    assert result_4087.raw == 0
    assert result_4087.value == 0
    assert result_4087.display_value == "0 kPa"

    results = pgn_1792.decode((0x0000010000000000).to_bytes(8, "big"))
    result_4087 = results[1]
    assert result_4087.raw == 1
    assert result_4087.value == 16
    assert result_4087.display_value == "16 kPa"

    results = pgn_1792.decode((0x0000FA0000000000).to_bytes(8, "big"))
    result_4087 = results[1]
    assert result_4087.value == 4000
    assert result_4087.display_value == "4000 kPa"


def test_it_decodes_multi_byte_spns(pgn_1792: PGN):
    results = pgn_1792.decode((0x0).to_bytes(8, "big"))
    result_4086 = results[0]
    assert result_4086.raw == 0
    assert result_4086.display_value == "0 kPa"

    results = pgn_1792.decode((0x0100000000000000).to_bytes(8, "big"))
    result_4087 = results[0]
    assert result_4087.value == 5
    assert result_4087.display_value == "5 kPa"

    results = pgn_1792.decode((64255).to_bytes(8, "little"))
    result_4087 = results[0]
    assert result_4087.value == 321275
    assert result_4087.display_value == "321275 kPa"

    results = pgn_1792.decode(0x1000000.to_bytes(8, "little"))
    result_4088 = results[2]
    assert result_4088.value == 5
    assert result_4088.display_value == "5 kPa"


class TestStartPositionValueExtractions:
    def test_it_extracts_whole_bytes(self):
        inp = (0x123456789ABCDEF0).to_bytes(8, "big")

        #  0x12 0x34 0x56 0x78 0x9a 0xbc 0xde 0xf0
        assert (0x12, 0) == extract_value_at_location(inp, "1", BL8)
        assert (0x34, 1) == extract_value_at_location(inp, "2", BL8)
        assert (0x56, 2) == extract_value_at_location(inp, "3", BL8)
        assert (0xDE, 6) == extract_value_at_location(inp, "7", BL8)
        assert (0xF0, 7) == extract_value_at_location(inp, "8", BL8)

        assert (0x3412, 1) == extract_value_at_location(
            inp, "1", BitLength(16)
        )
        assert (0x5634, 2) == extract_value_at_location(
            inp, "2", BitLength(16)
        )

        assert (0xF0DEBC9A78563412, 7) == extract_value_at_location(
            inp, "1", BitLength(64)
        )

    def test_it_extracts_fractions_of_a_byte(self):
        inp = (0x123456789ABCDEF0).to_bytes(8, "big")

        # 0xf0 0xde 0xbc 0x9a 0x78 0x56 0x34 0x12
        assert (0x2, 0) == extract_value_at_location(inp, "1", BitLength(4))
        assert (0x2, 0) == extract_value_at_location(inp, "1.1", BitLength(4))
        assert (0x1, 0) == extract_value_at_location(inp, "1.5", BitLength(4))
        assert (0xE, 6) == extract_value_at_location(inp, "7", BitLength(4))
        assert (0xD, 6) == extract_value_at_location(inp, "7.5", BitLength(4))

        # Within byte 0x12 => 0b00010010
        assert (0b1001, 0) == extract_value_at_location(
            inp, "1.2", BitLength(4)
        )
        assert (0b0100, 0) == extract_value_at_location(
            inp, "1.3", BitLength(4)
        )
        assert (0b010, 0) == extract_value_at_location(
            inp, "1.4", BitLength(3)
        )

    def test_it_extracts_across_byte_boundaries(self):
        inp = (0x123456789ABCDEF0).to_bytes(8, "big")
        # across byte 0x34 0x12 => 0b00110100 0b00010010
        assert (0b001101000001001, 1) == extract_value_at_location(
            inp, "1.2", BitLength(15)
        )
        assert (0b00110100000100, 1) == extract_value_at_location(
            inp, "1.3", BitLength(14)
        )
        assert (0b0011010000010, 1) == extract_value_at_location(
            inp, "1.4", BitLength(13)
        )
        assert (0b001101000001, 1) == extract_value_at_location(
            inp, "1.5", BitLength(12)
        )
        assert (0b00110100000, 1) == extract_value_at_location(
            inp, "1.6", BitLength(11)
        )
        assert (0b0011010000, 1) == extract_value_at_location(
            inp, "1.7", BitLength(10)
        )
        assert (0b001101000, 1) == extract_value_at_location(
            inp, "1.8", BitLength(9)
        )

    def test_it_extracts_across_bytes_where_last_position_is_on_bit_1(self):
        inp = (0x123456789ABCDEF0).to_bytes(8, "big")

        # Example PGN 39680, SPN 4181 start pos: 2.4-4
        # across byte 0x78 0x56 0x34 0x12 => 0b01111000 0b01010110 0b00110100 0b00010010
        #                                      X-------------------------X
        #         123456789012345678901
        assert (0b011110000101011000110, 3) == extract_value_at_location(
            inp, "2.4-4", BitLength(21)
        )

        # Examlpe PGN 64881, SPN 3795 start_pos: 4.7-5.1
        # Byte :     5    4    3    2    1
        #         0x9a 0x78 0x56 0x34 0x12 => 0b10011010 0b01111000 ...
        #                                              X----X
        assert (0b001, 4) == extract_value_at_location(
            inp, "4.7-5.1", BitLength(3)
        )

    def test_it_extracts_when_non_contiguous(self):
        """
        Non contigous specs mostly seem to have a comma in them, which separates
        the contiguous section from the "last few bits".

        For example:

        SPN 1214 is 19-bit and commonly has a spec that looks like:
        - "3-4, 5.6" in PGN 65226 ; or
        - "2-3, 4.6" in PGN 65229

        OR

        SPN 1596 is 12 bits, and has a spec like:
        - '1, 2.5' in PGN 54272

        SPN 1640 is 11 bits, and has a spec like:
        - '1, 2.6' in PGN 55552

        """
        inp = (0x123456789ABCDEF0).to_bytes(8, "big")

        # Example PGN 65226, SPN 1214,  19-bits, start_pos: '3-4, 5.6'
        # Byte :     5    4    3    2    1             5          4          3 ...
        #         0x9a 0x78 0x56 0x34 0x12 => 0b10011010 0b01111000 0b01010110 ...
        #                                       X-X        X-----------------X
        assert (0b1000111100001010110, 4) == extract_value_at_location(
            inp, "3-4, 5.6", BitLength(19)
        )

        # Example PGN 65229, SPN 1214,  19-bits, start_pos: '2-3, 4.6'
        # Byte :     4    3    2    1             4          3          2
        #         0x78 0x56 0x34 0x12 => 0b01111000 0b01010110 0b00110100 ...
        #                                  X-X        X-----------------X
        assert (0b0110101011000110100, 3) == extract_value_at_location(
            inp, "2-3, 4.6", BitLength(19)
        )

        # Example PGN 54272, SPN 1596,  12-bits, start_pos: '1, 2.5'
        # Byte :     2    1
        #         0x34 0x12 =>  0b00110100 0b00010010
        #                         X--X       X------X
        assert (0x312, 1) == extract_value_at_location(
            inp, "1, 2.5", BitLength(12)
        )

        # Example PGN 5552, SPN 1640,  12-bits, start_pos: '1, 2.6'
        # Byte :     2    1
        #         0x34 0x12 =>  0b00110100 0b00010010
        #                         X-X        X------X
        assert (0x112, 1) == extract_value_at_location(
            inp, "1, 2.6", BitLength(11)
        )

    def test_only_2_sections_are_allowed(self):
        with pytest.raises(UnsupportedLocationSpec) as e:
            extract_value_at_location(None, "1,2,3", None)
        assert (
            str(e.value)
            == "Should not have more than 2 sections in location spec"
        )

    def test_second_section_does_not_have_a_byte_range_spec(self):
        with pytest.raises(UnsupportedLocationSpec) as e:
            extract_value_at_location(None, "1,2-3", None)
        assert (
            str(e.value)
            == "Should not have a range specified in non-continuous part of location spec"
        )

    def test_wholly_contiguous_multi_byte_specs_should_have_the_second_addr_start_on_bit_1(
        self,
    ):
        with pytest.raises(UnsupportedLocationSpec) as e:
            extract_value_at_location(None, "1-2.3", BitLength(None, True))
        assert (
            str(e.value)
            == "Should not have a non-first-bit aligned location spec for contiguous bytes"
        )

    def test_fractional_bits_in_first_part(self):
        # Example PGN 10240, SPN 10786,  12-bits, start_pos: '2.5, 3'
        inp = (0x123456789ABCDEF0).to_bytes(8, "big")
        assert (0x0563, 2) == extract_value_at_location(
            inp, "2.5, 3", BitLength(12)
        )

        # Example PGN 10240, SPN 10787,  12-bits, start_pos: '4, 5.1'
        inp = (0x123456789ABCDEF0).to_bytes(8, "big")
        assert (0x0A78, 4) == extract_value_at_location(
            inp, "4, 5.1", BitLength(12)
        )


class TestStarDelimitedPGNs:
    @pytest.mark.parametrize(
        ["inp", "expected"],
        [
            ("hello*", "hello"),
            ("hello, world*", "hello, world"),
            ("x" * 1000, "x" * 200),
        ],
    )
    def test_it_decodes_pgn_65260(
        self, inp: str, expected: str, pgn_65260: PGN
    ):
        results = pgn_65260.decode(bytes(inp, "ascii"))
        assert results[0].value == expected


class TestPGNDecodesTextSPNs:
    def test_it_decodes_pgn_61445(self, pgn_61445):
        inp = [
            126,  # 1st gear
            0x31,
            0xD4,  # 54.321 .. 54321 = 0xD431
            129,  # 4th gear
            ord("D"),
            ord("1"),  # Requested D1
            ord("D"),
            ord("4"),  # Current D4
        ]  # yapf: disable
        results = pgn_61445.decode(bytes(inp))
        assert results[0].value == 1
        assert results[1].value == 54.321
        assert results[2].value == 4
        assert results[3].value == "D1"
        assert results[4].value == "D4"

    def test_it_decodes_pgn_43008(self, pgn_43008):
        # PGN 43008 Text Display -  has a "4 to N"
        inp = (
            bytes([3, 9, 0x12])
            + bytes("Abc123", "ascii")
            + bytes([0])
            + bytes("xyz", "ascii")
        )
        decoded = pgn_43008.decode(inp)

        assert decoded[0].raw == 3
        assert (
            decoded[0].value
            == "Overwrite display - The presently displayed information is to be overwritten with the transmitted information"
        )
        assert decoded[1].value == 0x12
        assert decoded[2].value == "Abc123"

    def test_it_others(self, pgn_1792: PGN):
        # PGN 64958 Transit Route - has a number of "X to Y" spns
        # PGN 64959 Transit Milepost - has a "2 to n" SPN
        # PGN 64792 - Collision Sensor Information -  has a bunch of packed strings
        results = pgn_1792.decode((0x0).to_bytes(8, "big"))
        result_4086 = results[0]
        assert result_4086.raw == 0
        assert result_4086.display_value == "0 kPa"


class TestStartPosition:
    def test_is_orders_basic_ints(self):
        a = OrderingRecord("1", None)
        b = OrderingRecord("2", None)
        assert sorted([b, a]) == [a, b]

    def test_it_orders_with_bit_offsets(self):
        a = OrderingRecord("1.1", None)
        b = OrderingRecord("1.2", None)
        c = OrderingRecord("2.1", None)
        assert sorted([b, c, a]) == [a, b, c]

    def test_it_orders_with_byte_range_hyphen(self):
        a = OrderingRecord("1-2", None)
        b = OrderingRecord("3-4", None)
        assert sorted([b, a]) == [a, b]

    def test_it_orders_with_byte_range_semicolon(self):
        a = OrderingRecord("1;2", None)
        b = OrderingRecord("3;4", None)
        assert sorted([b, a]) == [a, b]

    def test_it_orders_ints_by_numeric_value_not_string_ordering(self):
        a = OrderingRecord("1", None)
        b = OrderingRecord("02", None)
        assert sorted([b, a]) == [a, b]

    def test_it_orders_characters_with_without_case(self):
        a = OrderingRecord("a", None)
        b = OrderingRecord("B", None)
        assert sorted([b, a]) == [a, b]

    def test_it_orders_numbers_first(self):
        a = OrderingRecord("99", None)
        b = OrderingRecord("a", None)
        assert sorted([b, a]) == [a, b]


class TestTransportProtocol:
    @pytest.mark.parametrize(
        "control_byte", [0, 1, 2, 15, 18, 20, 31, 33, 254]
    )
    def test_it_decodes_pgn_60416_with_illegal_control_byte(
        self, control_byte: int, pgn_60416: PGN
    ):
        inp = control_byte.to_bytes(8, "little")
        results = pgn_60416.decode(inp)
        assert results[0].raw == control_byte
        assert results[0].value is None
        assert results[0].display_value == f"No encoding ({control_byte})"
        assert len(results) == 1

    def test_it_decodes_pgn_60416_CTS(self, pgn_60416: PGN):
        inp = (0x10F1F2F3F4F5F6F7).to_bytes(8, "big")
        results = pgn_60416.decode(inp)
        assert results[0].value == "Request to Send"
        assert len(results) == 5
        assert (
            str(results[1])
            == "SPN:2557: Total Message Size (TP.CM_RTS) = 62193"
        )
        assert (
            str(results[2])
            == "SPN:2558: Total Number of Packets (TP.CM_RTS) = 243"
        )
        assert (
            str(results[3])
            == "SPN:2559: Maximum Number of Packets (TP.CM_RTS) = 244"
        )
        assert (
            str(results[4])
            == "SPN:2560: Parameter Group Number of the packeted message (TP.CM_RTS) = 16250613"
        )

    def test_it_decodes_pgn_60416_RTS(self, pgn_60416: PGN):
        inp = (0x11F1F2F3F4F5F6F7).to_bytes(8, "big")
        results = pgn_60416.decode(inp)
        assert results[0].value == "Clear to Send"
        assert len(results) == 4
        assert (
            str(results[1])
            == "SPN:2561: Number of Packets that can be sent (TP.CM_CTS) = 241"
        )
        assert (
            str(results[2])
            == "SPN:2562: Next Packet Number to be sent (TP.CM_CTS) = 242"
        )
        assert (
            str(results[3])
            == "SPN:2563: Parameter Group Number of the packeted message (TP.CM_CTS) = 16250613"
        )

    def test_it_decodes_pgn_60416_EOM(self, pgn_60416: PGN):
        inp = (0x13F1F2F3F4F5F6F8).to_bytes(8, "big")
        results = pgn_60416.decode(inp)
        assert results[0].value == "End of Message ACK"
        assert len(results) == 4
        assert (
            str(results[1])
            == "SPN:2564: Total Message Size (TP.CM_EndofMsgACK) = 62193"
        )
        assert (
            str(results[2])
            == "SPN:2565: Total Number of Packets (TP.CM_EndofMsgACK) = 243"
        )
        assert (
            str(results[3])
            == "SPN:2566: Parameter Group Number of the packeted message (TP.CM_EndofMsgACK) = 16316149"
        )

    def test_it_decodes_pgn_60416_BAM(self, pgn_60416: PGN):
        inp = (0x20F1F2F3F4F5F6F8).to_bytes(8, "big")
        results = pgn_60416.decode(inp)
        assert results[0].value == "Broadcast Announce Message"
        assert len(results) == 4
        assert (
            str(results[1])
            == "SPN:2567: Total Message Size (TP.CM_BAM) = 62193"
        )
        assert (
            str(results[2])
            == "SPN:2568: Total Number of Packets (TP.CM_BAM) = 243"
        )
        assert (
            str(results[3])
            == "SPN:2569: Parameter Group Number of the packeted message (TP.CM_BAM) = 16316149"
        )

    def test_it_decodes_pgn_60416_ConnectionAbort(self, pgn_60416: PGN):
        inp = (0xFF01F2F3F4F5F6F8).to_bytes(8, "big")
        results = pgn_60416.decode(inp)
        assert results[0].value == "Connection Abort"
        assert len(results) == 3
        assert (
            str(results[1])
            == "SPN:2570: Connection Abort Reason = Already in one or more connection managed sessions and cannot support another."
        )
        assert (
            str(results[2])
            == "SPN:2571: Parameter Group Number of packeted message (TP.CM_Conn_Abort) = 16316149"
        )



# Copyright Andrew Dodd
from decoda import spec_provider
from decoda.main import UnknownReferenceError


def location_fmt(value, *args, **kwargs):
    if value >= 0xFF:
        return "Not available"
    axle = 1 + (0x0F & (value >> 4))
    tire = 1 + (0x0F & value)

    ordinals = {1: "st", 2: "nd", 3: "rd"}
    axle = f"{axle}{ordinals.get(axle, 'th')}"
    tire = f"{tire}{ordinals.get(tire, 'th')}"

    return f"{axle} axle, {tire} tire"


def refers_to_spn(value, *args, **kwargs):
    try:
        found = spec_provider.provide().SPNs.get_by_id(value)
        return f"SPN {value} - {found.name}"
    except UnknownReferenceError:
        return f"SPN {value}"


def conditionally_applies(
    value,
    *args,
    already_decoded=None,
    custom_args=None,
    non_custom_alternative=None,
    original_value=None,
    **kwargs,
):
    if (
        not custom_args
        or not custom_args.get("conditional_on_spn")
        or not custom_args.get("applies_if")
    ):
        raise ValueError("bad config")

    conditional_on_spn_id = custom_args["conditional_on_spn"]

    if not already_decoded:
        raise ValueError("no values decoded")

    try:
        decoded = next(
            d for d in already_decoded if d.id == conditional_on_spn_id
        )
    except StopIteration:
        raise ValueError(
            f"conditional on SPN {conditional_on_spn_id}, but not found"
        )

    if decoded.value != custom_args["applies_if"]:
        return None

    alternative_value = non_custom_alternative.decode_from_raw(
        value=original_value,
        already_decoded=already_decoded,
        custom_args=custom_args,
        non_custom_alternative=non_custom_alternative,
        **kwargs,
    )
    return alternative_value[0]["value"]



from decoda import *


def bytes_from_payload(payload):
    payload_len = len(payload)
    if payload_len % 2 != 0:
        raise Exception("incorrect length")
    if (payload_len // 2) < 8:
        raise Exception("insufficient bytes in payload")
    if payload_len // 2 > 1785:
        raise Exception("too many bytes in payload")
    ints = []
    while len(payload) > 0:
        ints.append(int(payload[:2], 16))
        payload = payload[2:]
    return bytes(ints)


def test_embeddeduse_peer_to_peer_example(spec: J1939Spec):
    # https://www.embeddeduse.com/2020/01/17/introduction-to-the-sae-j1939-standard/
    pgn = spec.PGNs.get_by_id(61184)
    payload = "0203029103000000"
    byte_payload = bytes_from_payload(payload)
    decoded = pgn.decode(byte_payload)

    spn_val = decoded[0]

    assert spn_val.value == bytes(
        [0x02, 0x03, 0x02, 0x91, 0x03, 0x00, 0x00, 0x00]
    )
    assert spn_val.display_value == payload


# TODO - Test the defragmentation



# Copyright Andrew Dodd
import attr

from .main import Address


def parts_from_pgn(pgn):
    edp = pgn >> 17
    dp = pgn >> 16
    pf = 0xFF & (pgn >> 8)
    ps = 0xFF & (pgn)
    return edp, dp, pf, ps


def parts_from_can_id(can_id):
    sa = 0xFF & can_id
    pgn = 0x3FFFF & (can_id >> 8)

    edp, dp, pf, ps = parts_from_pgn(pgn)
    da = Address.from_pf_and_ps(pf, ps)

    if pgn < 0xF000:
        pgn &= 0xFF00

    priority = 0x7 & (can_id >> (8 + 18))
    return priority, pgn, Address(sa), da


@attr.s(frozen=True)
class Message:
    priority = attr.ib()
    src_address = attr.ib()
    dst_address = attr.ib()
    pgn = attr.ib()
    decoded = attr.ib()


class Decoda:
    def __init__(self, spec, callback=None, error_handler=None):
        self.__spec__ = spec
        self._callback = callback if callback else lambda x: print(x)
        self._handle_error = (
            error_handler if error_handler else lambda x: print(x)
        )

    def get_callback(self):
        return self._callback

    def set_callback(self, callback):
        if not callback:
            raise ValueError("Callback must be valid")
        self._callback = callback

    def set_error_handler(self, error_handler):
        if not error_handler:
            raise ValueError("Error handler must be valid")
        self._handle_error = error_handler

    def handle_frame(self, can_id, payload):
        try:
            priority, pgn_id, sa, da = parts_from_can_id(can_id)
            pgn = self.__spec__.PGNs.get_by_id(pgn_id)
            decoded = pgn.decode(payload)

            self._callback(Message(priority, sa, da, pgn, decoded))
        except ValueError as e:
            self._handle_error(e)

    def handle_message(self, priority, sa, da, pgn_id, payload):
        try:
            pgn = self.__spec__.PGNs.get_by_id(pgn_id)
            decoded = pgn.decode(payload)
            self._callback(Message(priority, sa, da, pgn, decoded))
        except ValueError as e:
            self._handle_error(e)


class ConnectionManager:
    def __init__(self, decoda, defragmenting_error_callback):
        self._active_defrags = {}
        self._decoda = decoda
        self._callback = decoda.get_callback()
        self._handle_error = defragmenting_error_callback
        decoda.set_callback(self.handle_message)

    def handle_message(self, message):
        # First publish the message outwards
        self._callback(message)

        try:
            # Next, attempt to look for reassembly opportunities
            if message.pgn.id == 60416:  # TP - Connection Management
                control_spn = next(
                    value for value in message.decoded if value.id == 2556
                )

                if control_spn.value == "Clear to Send":
                    self._start_new_defragmenting(message)
                if control_spn.value == "Broadcast Announce Message":
                    self._start_new_defragmenting(message)

            if message.pgn.id == 60160:  # TP - Data Transfer
                self._handle_fragment(message)

        except Exception as e:
            print(e)
            raise

    def _start_new_defragmenting(self, message):
        to_from_pair = (message.src_address, message.dst_address)
        if to_from_pair in self._active_defrags:
            progress = self._active_defrags[to_from_pair]
            self._handle_error("incomplete defragmentation", progress)

        packet_count = next(
            value for value in message.decoded if value.id == 2561
        ).value
        next_packet_number = next(
            value for value in message.decoded if value.id == 2562
        ).value
        pgn = next(
            value for value in message.decoded if value.id == 2563
        ).value
        self._active_defrags[to_from_pair] = {
            "pgn": pgn,
            "packet_count": packet_count,
            "next_packet_number": next_packet_number,
            "fragments": [],
            "original_message": message,
        }

    def _handle_fragment(self, message):
        to_from_pair = (message.src_address, message.dst_address)
        if to_from_pair not in self._active_defrags:
            self._handle_error("fragment received without control", message)
            return

        pair_state = self._active_defrags[to_from_pair]
        seq_no = next(
            value for value in message.decoded if value.id == 2572
        ).value

        if seq_no != pair_state["next_packet_number"]:
            self._handle_error("fragment received out of order", message)
            return

        pair_state["next_packet_number"] = seq_no + 1
        fragment = next(
            value for value in message.decoded if value.id == 2573
        ).value
        pair_state["fragments"].append(fragment)

        if pair_state["next_packet_number"] > pair_state["packet_count"]:
            reassembled = bytes()
            for fragment in pair_state["fragments"]:
                reassembled += fragment
            original_msg = pair_state["original_message"]
            priority = original_msg.priority
            sa = original_msg.src_address
            da = original_msg.dst_address
            self._decoda.handle_message(
                priority, sa, da, pair_state["pgn"], reassembled
            )
            del self._active_defrags[to_from_pair]



# Copyright Andrew Dodd
import importlib
import json
import math
import os
from binascii import hexlify
from numbers import Number
from typing import Callable, Dict, Generic, List, Type, TypeVar

import attr

from decoda.exceptions import MissingBitLength, UnknownReferenceError
from decoda.main import (
    PGN,
    SPN,
    BitLength,
    ByteArrayValue,
    CustomFunctionValue,
    EncodedValue,
    IndustryGroup,
    Manufacturer,
    OrderingRecord,
    ScalarValue,
    TextValue,
)

T = TypeVar("T", covariant=True)


class Repo(Generic[T]):
    # TODO - Replace the passed cls with typing info
    def __init__(
        self, cls: Type[T], from_dict: Callable[[Dict], T], l, *args, **kwargs
    ):
        self.__cls = cls
        self.__lookup: Dict[int, T] = {
            int(r["id"]): from_dict(r, *args, **kwargs) for r in l
        }

    def get_by_id(self, id) -> T:
        try:
            return self.__lookup[id]
        except KeyError:
            raise UnknownReferenceError(
                "{} not found for id: {}".format(self.__cls.__name__, id)
            )


@attr.s(frozen=True)
class J1939Spec:
    Manufacturers: Repo[Manufacturer] = attr.ib()
    SPNs: Repo[SPN] = attr.ib()
    PGNs: Repo[PGN] = attr.ib()
    IndustryGroups: Repo[IndustryGroup] = attr.ib()

    def preferred_address_name(self, id, industry_group=0):
        if id == -1:
            return "Broadcast address"
        global_ig = self.IndustryGroups.get_by_id(0)
        global_name = global_ig.preferred_address_name(id)
        if id < 128:
            return global_name
        else:
            ig = self.IndustryGroups.get_by_id(industry_group)
        return ig.preferred_address_name(id) or global_name


def manufacturer_from_dict(d: Dict) -> Manufacturer:
    return Manufacturer(
        d["id"],
        d["name"],
        d.get("location", "No location"),
        d.get("status", "Unknown status"),
    )


def scalar_value_from_dict(d: Dict) -> ScalarValue:
    # unpack from the storage format
    units = d.get("units")
    data_range = d.get("data_range", None) or {}
    offset = d.get("offset", None)
    resolution = d.get("resolution", None)
    try:
        bit_length = BitLength(int(d.get("bit_length")))  # type: ignore
    except ValueError:
        raise MissingBitLength("ScalarValue SPN: {}".format(d["id"]))

    return ScalarValue(
        units,
        data_range.get("min", None),
        data_range.get("max", None),
        offset,
        resolution,
        bit_length,
    )


# Nod to DRF's module loading a la: https://github.com/encode/django-rest-framework/blob/8385ae42c06b8e68a714cb67b7f0766afe316883/rest_framework/settings.py
def import_from_string(val):
    """
    Attempt to import a class from a string representation.
    """
    try:
        # Nod to tastypie's use of importlib.
        parts = val.split(".")
        module_path, name = ".".join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, name)
    except (ImportError, AttributeError, ValueError) as e:
        msg = "Could not import '%s'. %s: %s." % (val, e.__class__.__name__, e)
        raise ImportError(msg)


def make_decoder_from_spn_dict(d):
    bl = d["bit_length"]
    variable = d.get("variable", False)

    try:
        # try to make an "integer" bit length
        bit_length = BitLength(int(bl), variable)
    except ValueError:
        # else just use what we found...downstream consumers will
        # check if this is ok for their needs
        bit_length = BitLength(bl, variable)

    if d.get("custom"):
        try:
            format_handler = import_from_string(d["custom"])
            non_custom = d.copy()
            del non_custom["custom"]
            alternative = make_decoder_from_spn_dict(non_custom)
            return CustomFunctionValue(
                format_handler, bit_length, alternative, d.get("custom_args")
            )
        except ImportError:
            pass

    if d.get("encodings"):
        return EncodedValue(d["encodings"], bit_length)
    elif (
        str(d.get("units", "")).lower() == "ascii"
        or d.get("delimiter", "")
        or d.get("length_spn")
    ):
        return TextValue(
            bit_length,
            delimiter=d.get("delimiter"),
            length_spn=d.get("length_spn"),
        )
    elif str(d.get("units", "")).lower() == "bytearray":
        return ByteArrayValue(bit_length)
    else:
        return scalar_value_from_dict(d)


def spn_from_dict(d: Dict) -> SPN:
    id = d["id"]
    name = d["name"]
    description = ""  # d.get("description", "")
    try:
        decoder = make_decoder_from_spn_dict(d)
    except ValueError as e:
        raise ValueError(f"Unable to make spn from {d}") from e
    return SPN(id, name, description, decoder)


def build_ordering_records(spn_list, spn_ref):
    spns = []
    for item in spn_list:
        start_pos = item["start_pos"]
        if not isinstance(start_pos, str):
            raise ValueError("Hmm...start pos is not a string")
        spns.append(OrderingRecord(start_pos, spn_ref.get_by_id(item["id"])))
    return sorted(spns)


def parse_length(v):
    try:
        return BitLength(int(v))
    except ValueError as ve:
        if "*" in v:
            return BitLength(None, True)
        if "variable" in v.lower():
            return BitLength(None, True)
        if v == "":
            return BitLength(8)
        print("Confusing length: {}".format(v))
        raise ve


def pgn_from_dict(d, spn_ref):
    return PGN(
        d["id"],
        d.get("name", d.get("label")),
        d["description"],
        d.get("transmission_rate"),
        parse_length(d["length"]),
        build_ordering_records(d["spns"], spn_ref),
        d.get("repeatable_spns", []),
        d.get("acronym"),
    )  # yapf: disable


def industry_group_from_dict(d, preferred_addresses):
    return IndustryGroup(
        d["id"], d["name"], preferred_addresses.get(d["id"], {})
    )


def convert_addresses(d):
    addresses = {}
    for k, v in d.items():
        if k == "Global":
            addresses[0] = {int(address): name for address, name in v.items()}
        else:
            k = int(k.replace("IG", ""))
            addresses[k] = {int(address): name for address, name in v.items()}
    return addresses


def load_from_file(filename, warn_of_errors=True):
    with open(filename) as f:
        spec = json.load(f)
    spns_in_error = []
    spn_specs = []
    for spn in spec["SPNs"]:
        try:
            spn_from_dict(spn)
            spn_specs.append(spn)
        except Exception as e:
            spns_in_error.append((spn, e))
            pass
    spns = Repo(SPN, spn_from_dict, spn_specs)

    pgns_in_error = []
    pgn_specs = []
    for pgn in spec["PGNs"]:
        try:
            pgn_from_dict(pgn, spns)
            pgn_specs.append(pgn)
        except Exception as e:
            pgns_in_error.append((pgn, e))
            pass

    pgns = Repo(PGN, pgn_from_dict, pgn_specs, spns)
    manufacturers = Repo(
        Manufacturer, manufacturer_from_dict, spec["Manufacturers"]
    )
    preferred_addresses = convert_addresses(spec["SourceAddresses"])
    industry_groups = Repo(
        IndustryGroup,
        industry_group_from_dict,
        spec["IndustryGroups"],
        preferred_addresses,
    )

    if warn_of_errors:
        if spns_in_error:
            spn_ids = sorted(
                list(set(spn.get("id") for spn, _ in spns_in_error))
            )
            print(f"SPN IDs with errors: {spn_ids}")
        if pgns_in_error:
            pgn_ids = sorted(
                list(set(pgn.get("id") for pgn, _ in pgns_in_error))
            )
            print(f"PGN IDs with errors: {pgn_ids}")

    return J1939Spec(manufacturers, spns, pgns, industry_groups)


# This is a rather ugly technique for now...just want to supply a lazy-loaded
# spec.
class LoadOnFirstAccess:
    def __init__(self):
        self.j1939 = None

    def provide(self):
        if not self.j1939:
            SPEC_PATH = os.environ.get("J1939_SPEC_PATH", "./decoda_spec.json")
            self.j1939 = load_from_file(SPEC_PATH)
        return self.j1939


spec_provider = LoadOnFirstAccess()



import pytest

from decoda import *


def test_spn899(spec):
    sut = spec.SPNs.get_by_id(899)
    test_input = {
        b"\x00": "Low idle governor/no request (default mode)",
        b"\x01": "Accelerator pedal/operator selection",
        b"\x02": "Cruise control",
        b"\x03": "PTO governor",
        b"\x04": "Road speed governor ",
        b"\x05": "ASR control ",
        b"\x06": "Transmission control ",
        b"\x07": "ABS control",
        b"\x08": "Torque limiting",
        b"\x09": "High speed governor ",
        b"\x0a": "Braking system ",
        b"\x0b": "Remote accelerator ",
        b"\x0c": "Service procedure ",
        b"\x0d": "Not defined",
        b"\x0e": "Other",
        b"\x0f": "Not available",
    }
    for encoded, expected in test_input.items():
        decoded, _ = sut.decode(encoded, "1", False, -1)
        assert decoded.value == expected


def test_location_spns(spec):
    for spn in range(927, 929):
        sut = spec.SPNs.get_by_id(spn)
        decoded, _ = sut.decode(b"\x23", "1", False, -1)
        assert decoded.value == "3rd axle, 4th tire"


def custom_fmt_func(value, *args, **kwargs):
    return {0: "ZERO", 1: "Een", 2: "Deux"}.get(value, value)


# yapf: disable
@pytest.mark.parametrize(
    ['input_value', 'raw', 'value', 'display_value'], [
    [      b"\x00",    0,   "ZERO", "ZERO (0)"],
    [      b"\x01",    1,    "Een", "Een (1)"],
    [      b"\x02",    2,   "Deux", "Deux (2)"],
    [      b"\x03",    3,        3, "3 (3)"],
])
# yapf: enable
def test_any_function_can_be_defined(input_value, value, raw, display_value):
    sut = make_decoder_from_spn_dict(
        {"bit_length": 8, "custom": "tests.test_adhoc_spns.custom_fmt_func"}
    )
    decoded, _ = sut.decode_from_raw(input_value, "1", False, -1)
    assert decoded["value"] == value
    assert decoded["raw"] == raw
    assert decoded["display_value"] == display_value


def test_it_falls_back_if_custom_function_not_found():
    sut = make_decoder_from_spn_dict(
        {
            "bit_length": 8,
            "custom": "does_not_exist",
            "encodings": {"0": "Zilch", "1-255": "Better than nothing"},
        }
    )
    decoded, _ = sut.decode_from_raw(b"\x00", "1", False, -1)
    assert decoded["value"] == "Zilch"
    decoded, _ = sut.decode_from_raw(b"\x1d", "1", False, -1)
    assert decoded["value"] == "Better than nothing"


@pytest.mark.parametrize(
    "spn", [4307, 5150, 5151, 5155, 5227] + list(range(5159, 5194))
)
def test_tractor_limit_statuses(spn, spec):
    sut = spec.SPNs.get_by_id(spn)
    decoded, _ = sut.decode(b"\x00", "1", False, -1)
    assert decoded.value == "Not limited"
    decoded, _ = sut.decode(b"\x03", "1", False, -1)
    assert (
        decoded.value
        == "Limited low (only higher command values result in a change)"
    )


@pytest.mark.parametrize("spn", [5152, 5153, 5154, 5156, 5157, 5158])
def test_tractor_request_status(spn, spec):
    sut = spec.SPNs.get_by_id(spn)
    decoded, _ = sut.decode(b"\x00", "1", False, -1)
    assert (
        decoded.value
        == "External request accepted. No subsequent operator intervention"
    )
    decoded, _ = sut.decode(b"\x03", "1", False, -1)
    assert decoded.value == "Not available (parameter not supported)"


def test_failure_mode_identifier(spec):
    sut = spec.SPNs.get_by_id(1215)
    decoded, _ = sut.decode(b"\x00", "1", False, -1)
    assert (
        decoded.value
        == "Data Valid But Above Normal Operational Range - Most Severe Level"
    )
    decoded, _ = sut.decode(b"\x03", "1", False, -1)
    assert decoded.value == "Voltage Above Normal, Or Shorted To High Source"
    decoded, _ = sut.decode(b"\x16", "1", False, -1)
    assert decoded.value == "Reserved For SAE Assignment"
    decoded, _ = sut.decode(b"\x1e", "1", False, -1)
    assert decoded.value == "Reserved For SAE Assignment"
    decoded, _ = sut.decode(b"\x1f", "1", False, -1)
    assert decoded.value == "Condition Exists"



#
# Copyright (c) 2019 National Motor Freight Traffic Association Inc. All Rights Reserved.
# See the file "LICENSE" for the full license governing this code.
#

import argparse
import functools
import itertools
import json
import operator
import re
import sys
from collections import OrderedDict

import asteval
import defusedxml
import unidecode
import xlrd
from defusedxml.common import EntitiesForbidden

# import pretty_j1939.describe

ENUM_SINGLE_LINE_RE = r"[ ]*([0-9bxXA-F]+)[ ]*[-=:]?(.*)"
ENUM_RANGE_LINE_RE = (
    r"[ ]*([0-9bxXA-F]+)[ ]*(\-|to|thru)[ ]*([0-9bxXA-F]+)[ ]+[-=:]?(.*)"
)

parser = argparse.ArgumentParser()
parser.add_argument(
    "-f",
    "--digital_annex_xls",
    type=str,
    required=True,
    action="append",
    default=[],
    nargs="+",
    help="the J1939 Digital Annex .xls excel file used as input",
)
parser.add_argument(
    "-w",
    "--write-json",
    type=str,
    default="-",
    help="where to write the output. defaults to stdout",
)


class J1939daConverter:
    def __init__(self, digital_annex_xls_list):
        defusedxml.defuse_stdlib()
        self.j1939db = OrderedDict()
        self.digital_annex_xls_list = list(
            map(
                lambda da: self.secure_open_workbook(
                    filename=da, on_demand=True
                ),
                digital_annex_xls_list,
            )
        )

    @staticmethod
    def secure_open_workbook(**kwargs):
        try:
            return xlrd.open_workbook(**kwargs)
        except EntitiesForbidden:
            raise ValueError("Please use an excel file without XEE")

    @staticmethod
    # returns a string of number of bits, or 'Variable', or ''
    def get_pgn_data_len(contents):
        if type(contents) is float:
            return str(int(contents))
        elif (
            "bytes" not in contents.lower()
            and "variable" not in contents.lower()
        ):
            return str(contents)
        elif "bytes" in contents.lower():
            return str(int(contents.split(" ")[0]) * 8)
        elif "variable" in contents.lower():
            return "Variable"
        elif contents.strip() == "":
            return ""
        raise ValueError('unknown PGN Length "%s"' % contents)

    @staticmethod
    # returns an int number of bits, or 'Variable'
    def get_spn_len(contents):
        if (
            "to" in contents.lower()
            or contents.strip() == ""
            or "variable" in contents.lower()
        ):
            return "Variable"
        elif re.match(r"max [0-9]+ bytes", contents):
            return "Variable"
        elif "byte" in contents.lower():
            return int(contents.split(" ")[0]) * 8
        elif "bit" in contents.lower():
            return int(contents.split(" ")[0])
        elif re.match(r"^[0-9]+$", contents):
            return int(contents)
        raise ValueError('unknown SPN Length "%s"' % contents)

    @staticmethod
    # returns a single-byte delimiter or None
    def get_spn_delimiter(contents):
        if "delimiter" in contents.lower():
            if "*" in contents:
                return b"*"
            elif "NULL" in contents:
                return b"\x00"
            else:
                raise ValueError('unknown SPN delimiter "%s"' % contents)
        else:
            return None

    @staticmethod
    def just_numeric_expr(contents):
        contents = re.sub(
            r"[^0-9\.\-/]", "", contents
        )  # remove all but number and '.'
        contents = re.sub(
            r"[/-]+[ ]*$", "", contents
        )  # remove trailing '/' or '-' that are sometimes left
        return contents

    @staticmethod
    def get_spn_units(contents, raw_spn_resolution):
        norm_contents = unidecode.unidecode(contents).lower().strip()
        raw_spn_resolution = (
            unidecode.unidecode(raw_spn_resolution).lower().strip()
        )
        if norm_contents == "":
            if "states" in raw_spn_resolution:
                norm_contents = "bit"
            elif "bit-mapped" in raw_spn_resolution:
                norm_contents = "bit-mapped"
            elif "binary" in raw_spn_resolution:
                norm_contents = "binary"
            elif "ascii" in raw_spn_resolution:
                norm_contents = "ascii"
        return norm_contents

    @staticmethod
    # returns a float in X per bit or int(0)
    def get_spn_resolution(contents):
        norm_contents = unidecode.unidecode(contents).lower()
        if (
            "0 to 255 per byte" in norm_contents
            or " states" in norm_contents
            or norm_contents == "data specific"
        ):
            return 1.0
        elif (
            "bit-mapped" in norm_contents
            or "binary" in norm_contents
            or "ascii" in norm_contents
            or "not defined" in norm_contents
            or "variant determined" in norm_contents
            or "7 bit iso latin 1 characters" in norm_contents
            or contents.strip() == ""
        ):
            return int(0)
        elif "per bit" in norm_contents or "/bit" in norm_contents:
            expr = J1939daConverter.just_numeric_expr(norm_contents)
            return J1939daConverter.asteval_eval(expr)
        elif "bit" in norm_contents and "/" in norm_contents:
            left, right = contents.split("/")
            left = J1939daConverter.just_numeric_expr(left)
            right = J1939daConverter.just_numeric_expr(right)
            return J1939daConverter.asteval_eval("(%s)/(%s)" % (left, right))
        elif (
            "microsiemens/mm" in norm_contents
            or "usiemens/mm" in norm_contents
            or "kw/s" in norm_contents
        ):  # special handling for this weirdness
            return float(contents.split(" ")[0])
        raise ValueError('unknown spn resolution "%s"' % contents)

    @staticmethod
    def asteval_eval(expr):
        interpreter = asteval.Interpreter()
        ret = interpreter(expr)
        if len(interpreter.error) > 0:
            raise interpreter.error[0]
        return ret

    @staticmethod
    # returns a float in 'units' of the SPN or int(0)
    def get_spn_offset(contents):
        norm_contents = unidecode.unidecode(contents).lower()
        if (
            "manufacturer defined" in norm_contents
            or "not defined" in norm_contents
            or contents.strip() == ""
        ):
            return int(0)
        else:
            first = J1939daConverter.just_numeric_expr(contents)
            return J1939daConverter.asteval_eval(first)

    @staticmethod
    # returns a pair of floats (low, high) in 'units' of the SPN or (-1, -1) for undefined operational ranges
    def get_operational_hilo(contents, units, spn_length):
        norm_contents = contents.lower()
        if contents.strip() == "" and units.strip() == "":
            if type(spn_length) is int:
                return 0, 2**spn_length - 1
            else:
                return -1, -1
        elif (
            "manufacturer defined" in norm_contents
            or "bit-mapped" in norm_contents
            or "not defined" in norm_contents
            or "variant determined" in norm_contents
            or contents.strip() == ""
        ):
            return -1, -1
        elif " to " in norm_contents:
            left, right = norm_contents.split(" to ")[0:2]
            left = J1939daConverter.just_numeric_expr(left)
            right = J1939daConverter.just_numeric_expr(right)

            range_units = norm_contents.split(" ")
            range_units = range_units[len(range_units) - 1]
            lo = float(J1939daConverter.asteval_eval(left))
            hi = float(J1939daConverter.asteval_eval(right))
            if range_units == "km" and units == "m":
                return lo * 1000, hi * 1000
            else:
                return lo, hi
        raise ValueError(
            'unknown operational range from "%s","%s"' % (contents, units)
        )

    @staticmethod
    # return a list of int of the start bits ([some_bit_pos] or [some_bit_pos,some_other_bit_pos]) of the SPN; or [
    # -1] (if unknown or variable).
    def get_spn_start_bit(contents):
        norm_contents = contents.lower()

        if ";" in norm_contents:  # special handling for e.g. '0x00;2'
            return [-1]

        # Explanation of multi-startbit (from J4L): According to 1939-71, "If the data length is larger than 1 byte
        # or the data spans a byte boundary, then the Start Position consists of two numerical values separated by a
        # comma or dash." Therefore , and - may be treated in the same way, multi-startbit. To account for
        # multi-startbit we will introduce the following: 1> an SPN position is now a pair of bit positions (R,S),
        # where S = None if not multibit 2> the SPN length is now a pair (Rs, Ss), where Ss = None if not multibit,
        # else net Rs = (S - R + 1) and Ss = (Length - Rs)

        delim = ""
        firsts = [norm_contents]
        if "," in norm_contents:
            delim = ","
        if "-" in norm_contents:
            delim = "-"
        elif " to " in norm_contents:
            delim = " to "

        if len(delim) > 0:
            firsts = norm_contents.split(delim)

        if any(re.match(r"^[a-z]\+[0-9]", first) for first in firsts):
            return [-1]

        firsts = [
            J1939daConverter.just_numeric_expr(first) for first in firsts
        ]
        if any(first.strip() == "" for first in firsts):
            return [-1]

        pos_pair = []
        for first in firsts:
            if "." in first:
                byte_index, bit_index = list(map(int, first.split(".")))
            else:
                bit_index = 1
                byte_index = int(first)
            pos_pair.append((byte_index - 1) * 8 + (bit_index - 1))

        return pos_pair

    @staticmethod
    def is_enum_line(line):
        if line.lower().startswith("bit state"):
            return True
        elif re.match(r"^[ ]*[0-9][0-9bxXA-F\-:]*[ ]+[^ ]+", line):
            return True
        return False

    @staticmethod
    def get_enum_lines(description_lines):
        enum_lines = list()

        def add_enum_line(test_line):
            test_line = re.sub(
                r"(Bit States|Bit State)", "", test_line, flags=re.IGNORECASE
            )
            if any(
                e in test_line
                for e in [
                    ":  Tokyo",
                    " SPN 8846 ",
                    " SPN 8842 ",
                    " SPN 3265 ",
                    " SPN 3216 ",
                    "13 preprogrammed intermediate ",
                    "3 ASCII space characters",
                ]
            ):
                return False
            enum_lines.append(test_line)
            return True

        any_found = False
        for line in description_lines:
            if J1939daConverter.is_enum_line(line):
                if any_found:
                    add_enum_line(line)
                else:
                    if J1939daConverter.match_single_enum_line(
                        line
                    ):  # special handling: first enum must use single assignment
                        any_found = add_enum_line(line)

        return enum_lines

    @staticmethod
    def is_enum_lines_binary(enum_lines_only):
        all_ones_and_zeroes = True
        for line in enum_lines_only:
            first = J1939daConverter.match_single_enum_line(line).groups()[0]
            if re.sub(r"[^10b]", "", first) != first:
                all_ones_and_zeroes = False
                break

        return all_ones_and_zeroes

    @staticmethod
    # returns a pair of inclusive, inclusive range boundaries or None if this line is not a range
    def get_enum_line_range(line):
        match = re.match(ENUM_RANGE_LINE_RE, line)
        if match:
            groups = match.groups()
            if re.match(r"[01b]", groups[0]) and not re.match(
                r"[01b]", groups[2]
            ):
                return None
            return groups[0], groups[2]
        else:
            return None

    @staticmethod
    def match_single_enum_line(line):
        line = re.sub(r"[ ]+", " ", line)
        line = re.sub(r"[ ]?\-\-[ ]?", " = ", line)
        return re.match(ENUM_SINGLE_LINE_RE, line)

    @staticmethod
    # returns the description part (just that part) of an enum line
    def get_enum_line_description(line):
        line = re.sub(r"[ ]+", " ", line)
        line = re.sub(r"[ ]?\-\-[ ]?", " = ", line)
        match = re.match(ENUM_RANGE_LINE_RE, line)
        if match:
            line = match.groups()[-1]
        else:
            match = J1939daConverter.match_single_enum_line(line)
            if match:
                line = match.groups()[-1]
        line = line.strip()
        line = line.lower()
        line = line.replace("sae", "SAE").replace("iso", "ISO")
        return line

    @staticmethod
    def create_bit_object_from_description(spn_description, bit_object):
        description_lines = spn_description.splitlines()
        enum_lines = J1939daConverter.get_enum_lines(description_lines)
        is_binary = J1939daConverter.is_enum_lines_binary(enum_lines)

        for line in enum_lines:
            line = re.sub(
                r"^(\d+)\s*-\s*(?!\=)", r"\1 = ", line
            )  # replace hyphen with = when in form "123 - "
            enum_description = J1939daConverter.get_enum_line_description(line)

            range_boundaries = J1939daConverter.get_enum_line_range(line)
            if range_boundaries is not None:
                if is_binary:
                    first = re.sub(r"b", "", range_boundaries[0])
                    first_val = int(first, base=2)
                    second = re.sub(r"b", "", range_boundaries[1])
                    second_val = int(second, base=2)
                elif "x" in range_boundaries[0].lower():
                    first_val = int(range_boundaries[0], base=16)
                    second_val = int(range_boundaries[1], base=16)
                else:
                    first_val = int(range_boundaries[0], base=10)
                    second_val = int(range_boundaries[1], base=10)

                for i in range(first_val, second_val + 1):
                    bit_object.update(({str(i): enum_description}))
            else:
                first = re.match(r"[ ]*([0-9bxXA-F]+)", line).groups()[0]

                if is_binary:
                    first = re.sub(r"b", "", first)
                    val = str(int(first, base=2))
                elif "x" in first.lower():
                    val = str(int(first, base=16))
                else:
                    val = str(int(first, base=10))

                bit_object.update(({val: enum_description}))

    @staticmethod
    def is_spn_likely_bitmapped(spn_description):
        return (
            len(J1939daConverter.get_enum_lines(spn_description.splitlines()))
            > 2
        )

    def process_spns_and_pgns_tab(self, sheet):
        self.j1939db.update({"J1939PGNdb": OrderedDict()})
        j1939_pgn_db = self.j1939db.get("J1939PGNdb")
        self.j1939db.update({"J1939SPNdb": OrderedDict()})
        j1939_spn_db = self.j1939db.get("J1939SPNdb")
        self.j1939db.update({"J1939BitDecodings": OrderedDict()})
        j1939_bit_decodings = self.j1939db.get("J1939BitDecodings")

        # check for SPNs in multiple PNGs
        spn_factcheck_map = dict()

        header_row, header_row_num = self.get_header_row(sheet)
        pgn_col = self.get_any_header_column(header_row, "PGN")
        spn_col = self.get_any_header_column(header_row, "SPN")
        acronym_col = self.get_any_header_column(
            header_row, ["ACRONYM", "PG_ACRONYM"]
        )
        pgn_label_col = self.get_any_header_column(
            header_row, ["PARAMETER_GROUP_LABEL", "PG_LABEL"]
        )
        pgn_data_length_col = self.get_any_header_column(
            header_row, ["PGN_DATA_LENGTH", "PG_DATA_LENGTH"]
        )
        transmission_rate_col = self.get_any_header_column(
            header_row, "TRANSMISSION_RATE"
        )
        spn_position_in_pgn_col = self.get_any_header_column(
            header_row, ["SPN_POSITION_IN_PGN", "SP_POSITION_IN_PG"]
        )
        spn_name_col = self.get_any_header_column(
            header_row, ["SPN_NAME", "SP_LABEL"]
        )
        offset_col = self.get_any_header_column(header_row, "OFFSET")
        data_range_col = self.get_any_header_column(header_row, "DATA_RANGE")
        resolution_col = self.get_any_header_column(
            header_row, ["RESOLUTION", "SCALING"]
        )
        spn_length_col = self.get_any_header_column(
            header_row, ["SPN_LENGTH", "SP_LENGTH"]
        )
        units_col = self.get_any_header_column(header_row, ["UNITS", "UNIT"])
        operational_range_col = self.get_any_header_column(
            header_row, "OPERATIONAL_RANGE"
        )
        spn_description_col = self.get_any_header_column(
            header_row, ["SPN_DESCRIPTION", "SP_DESCRIPTION"]
        )

        for i in range(header_row_num + 1, sheet.nrows):
            row = sheet.row_values(i)
            pgn = row[pgn_col]
            if pgn == "":
                continue

            pgn_label = str(int(pgn))

            spn = row[spn_col]

            if not j1939_pgn_db.get(pgn_label) is None:
                # TODO assert that PGN values haven't changed across multiple SPN rows
                pass
            else:
                pgn_object = OrderedDict()

                pgn_data_len = self.get_pgn_data_len(row[pgn_data_length_col])

                pgn_object.update(
                    {"Label": unidecode.unidecode(row[acronym_col])}
                )
                pgn_object.update(
                    {"Name": unidecode.unidecode(row[pgn_label_col])}
                )
                pgn_object.update({"PGNLength": pgn_data_len})
                pgn_object.update(
                    {"Rate": unidecode.unidecode(row[transmission_rate_col])}
                )
                pgn_object.update({"SPNs": list()})
                pgn_object.update({"SPNStartBits": list()})
                pgn_object.update({"Temp_SPN_Order": list()})

                j1939_pgn_db.update({pgn_label: pgn_object})

            if pretty_j1939.describe.is_transport_pgn(
                int(pgn)
            ):  # skip all SPNs for transport PGNs
                continue

            if not spn == "":
                if spn_factcheck_map.get(spn, None) is None:
                    spn_factcheck_map.update(
                        {
                            spn: [
                                pgn,
                            ]
                        }
                    )
                else:
                    spn_list = spn_factcheck_map.get(spn)
                    spn_list.append(spn)
                    spn_factcheck_map.update({spn: spn_list})

                spn_label = str(int(spn))
                spn_object = OrderedDict()

                spn_length = self.get_spn_len(row[spn_length_col])
                if type(spn_length) == str and spn_length.startswith(
                    "Variable"
                ):
                    spn_delimiter = self.get_spn_delimiter(row[spn_length_col])
                else:
                    spn_delimiter = None

                spn_resolution = self.get_spn_resolution(row[resolution_col])
                spn_units = self.get_spn_units(
                    row[units_col], row[resolution_col]
                )
                data_range = unidecode.unidecode(row[data_range_col])
                low, high = self.get_operational_hilo(
                    data_range, spn_units, spn_length
                )

                spn_name = unidecode.unidecode(row[spn_name_col])
                operational_range = unidecode.unidecode(
                    row[operational_range_col]
                )
                spn_offset = self.get_spn_offset(row[offset_col])

                spn_object.update({"DataRange": data_range})
                spn_object.update({"Name": spn_name})
                spn_object.update({"Offset": spn_offset})
                spn_object.update({"OperationalHigh": high})
                spn_object.update({"OperationalLow": low})
                spn_object.update({"OperationalRange": operational_range})
                spn_object.update({"Resolution": spn_resolution})
                spn_object.update({"SPNLength": spn_length})
                if spn_delimiter is not None:
                    spn_object.update(
                        {"Delimiter": "0x%s" % spn_delimiter.hex()}
                    )
                spn_object.update({"Units": spn_units})

                existing_spn = j1939_spn_db.get(str(int(spn)))
                if existing_spn is not None and not existing_spn == spn_object:
                    print(
                        "Warning: changed details of SPN %s:\n %s vs previous:\n %s"
                        % (spn, existing_spn, spn_object),
                        file=sys.stderr,
                    )
                else:
                    j1939_spn_db.update({spn_label: spn_object})

                # record SPN position-in-PGN ('StartBit') in the PGN structure along with the list of SPNs -- or skip
                # this SPN
                try:
                    spn_position_contents = row[spn_position_in_pgn_col]
                    spn_startbit_inpgn = self.get_spn_start_bit(
                        spn_position_contents
                    )
                    if (
                        spn_label == "5998"
                        and spn_position_contents.strip() == "4.4"
                    ):  # bug in 201311 DA
                        spn_startbit_inpgn = self.get_spn_start_bit("4.5")
                    elif (
                        spn_label == "3036"
                        and spn_position_contents.strip() == "6-8.6"
                    ):  # bug in 201311 DA
                        spn_startbit_inpgn = self.get_spn_start_bit("6-7,8.6")
                    elif (
                        spn_label == "6062"
                        and spn_position_contents.strip() == "4.4"
                    ):  # bug in 201311 DA
                        spn_startbit_inpgn = self.get_spn_start_bit("4.5")
                    elif (
                        spn_label == "6030"
                        and spn_position_contents.strip() == "4.4"
                    ):  # bug in 201311 DA
                        spn_startbit_inpgn = self.get_spn_start_bit("4.5")

                    if spn_startbit_inpgn == [-1]:
                        spn_order_inpgn = spn_position_contents.strip()
                    else:
                        spn_order_inpgn = spn_startbit_inpgn
                except ValueError:
                    continue

                if (
                    spn_label == "6610" or spn_label == "6815"
                ):  # bug in PGN map in 201311 DA
                    continue

                # Back to PGN processing

                j1939_pgn_db.get(pgn_label).get("SPNs").append(int(spn))
                # TODO strip consecutive startbits e.g. '[8, 16, 24]' for a 24bit val should be just '8'
                j1939_pgn_db.get(pgn_label).get("SPNStartBits").append(
                    [int(s) for s in spn_startbit_inpgn]
                )
                # the Temp_SPN_Order list will be deleted later
                j1939_pgn_db.get(pgn_label).get("Temp_SPN_Order").append(
                    spn_order_inpgn
                )

                # If there is a bitfield/enum described in this row, then create a separate object describing the states
                spn_description = unidecode.unidecode(row[spn_description_col])
                if row[units_col] == "bit" or self.is_spn_likely_bitmapped(
                    spn_description
                ):
                    bit_object = OrderedDict()
                    self.create_bit_object_from_description(
                        spn_description, bit_object
                    )
                    if len(bit_object) > 0:
                        j1939_bit_decodings.update({spn_label: bit_object})

        # Clean-ups are needed. The next steps are to do:
        # 1. sort SPN lists in PGNs by the Temp_SPN_Order
        # 2. fix the starting sequence of -1 startbits in PGNs with fixed-len SPNs mapped
        # 3. fix incorrectly variable-len SPNs in a sequence known startbits
        # 4. remove any SPN maps that have variable-len, no-delimiter SPNs in a PGN with >1 SPN mapped
        # 5. remove Temp_SPN_Order
        # 6. remove zero-len startbits arrays

        # * sort SPN lists in PGNs by the Temp_SPN_Order
        self.sort_spns_by_order(j1939_pgn_db)

        # * fix the starting sequence of -1 startbits in PGNs with fixed-len SPNs mapped
        self.remove_startbitsunknown_spns(j1939_pgn_db, j1939_spn_db)

        # * fix incorrectly variable-len SPNs in a sequence known startbits
        self.fix_omittedlen_spns(j1939_pgn_db, j1939_spn_db)

        # * remove any SPN maps that have variable-len, no-delimiter SPNs in a PGN with >1 SPN mapped
        self.remove_underspecd_spns(j1939_pgn_db, j1939_spn_db)

        # * remove Temp_SPN_Order
        for pgn, pgn_object in j1939_pgn_db.items():
            pgn_object.pop("Temp_SPN_Order")

        # * remove zero-len startbits arrays
        for pgn, pgn_object in j1939_pgn_db.items():
            spn_list = pgn_object.get("SPNs")
            if len(spn_list) == 0:
                pgn_object.pop("SPNStartBits")

        return

    def get_any_header_column(self, header_row, header_texts):
        if not isinstance(header_texts, list):
            header_texts = [header_texts]
        for t in header_texts:
            try:
                return header_row.index(t)
            except ValueError:
                continue
        return -1

    def get_header_row(self, sheet):
        header_row_num = self.lookup_header_row(sheet)

        header_row = sheet.row_values(header_row_num)
        header_row = list(map(lambda x: x.upper(), header_row))
        header_row = list(map(lambda x: x.replace(" ", "_"), header_row))
        return header_row, header_row_num

    def lookup_header_row(self, sheet):
        if sheet.row_values(0)[3].strip() == "":
            return 3
        else:
            return 0

    @staticmethod
    def fix_omittedlen_spns(j1939_pgn_db, j1939_spn_db):
        modified_spns = dict()
        for pgn, pgn_object in j1939_pgn_db.items():
            spn_list = pgn_object.get("SPNs")
            spn_startbit_list = pgn_object.get("SPNStartBits")
            spn_order_list = pgn_object.get("Temp_SPN_Order")

            spn_in_pgn_list = list(
                zip(spn_list, spn_startbit_list, spn_order_list)
            )
            if J1939daConverter.all_spns_positioned(spn_startbit_list):
                for i in range(0, len(spn_in_pgn_list) - 1):
                    here_startbit = int(spn_in_pgn_list[i][1][0])
                    next_startbit = int(spn_in_pgn_list[i + 1][1][0])
                    calced_spn_length = next_startbit - here_startbit
                    here_spn = spn_in_pgn_list[i][0]

                    if calced_spn_length == 0:
                        print(
                            "Warning: calculated zero-length SPN %s in PGN %s"
                            % (here_spn, pgn),
                            file=sys.stderr,
                        )
                        continue
                    else:
                        spn_obj = j1939_spn_db.get(str(here_spn))
                        current_spn_length = spn_obj.get("SPNLength")
                        if J1939daConverter.is_length_variable(
                            current_spn_length
                        ):
                            spn_obj.update({"SPNLength": calced_spn_length})
                            modified_spns.update({here_spn: True})
                        elif (
                            calced_spn_length < current_spn_length
                            and modified_spns.get(here_spn) is None
                        ):
                            print(
                                "Warning: calculated length for SPN %s (%d) in PGN %s differs from existing SPN "
                                "length %s"
                                % (
                                    here_spn,
                                    calced_spn_length,
                                    pgn,
                                    current_spn_length,
                                ),
                                file=sys.stderr,
                            )

    @staticmethod
    def is_length_variable(spn_length):
        return type(spn_length) is str and spn_length.startswith("Variable")

    @staticmethod
    def remove_startbitsunknown_spns(j1939_pgn_db, j1939_spn_db):
        for pgn, pgn_object in j1939_pgn_db.items():
            spn_list = pgn_object.get("SPNs")
            if len(spn_list) > 1:
                spn_list = pgn_object.get("SPNs")
                spn_startbit_list = pgn_object.get("SPNStartBits")
                spn_order_list = pgn_object.get("Temp_SPN_Order")

                spn_in_pgn_list = list(
                    zip(spn_list, spn_startbit_list, spn_order_list)
                )
                for i in range(0, len(spn_in_pgn_list)):
                    here_startbit = int(spn_in_pgn_list[i][1][0])
                    prev_spn = spn_in_pgn_list[i - 1][0]
                    prev_spn_obj = j1939_spn_db.get(str(prev_spn))
                    prev_spn_len = prev_spn_obj.get("SPNLength")
                    if (
                        here_startbit == -1
                        and not J1939daConverter.is_length_variable(
                            prev_spn_len
                        )
                    ):
                        if (i - 1) == 0:  # special case for the first field
                            prev_startbit = 0
                            here_startbit = prev_spn_len
                            prev_tuple = list(spn_in_pgn_list[i - 1])
                            prev_tuple[1] = [prev_startbit]
                            spn_in_pgn_list[i - 1] = tuple(prev_tuple)
                        else:
                            prev_startbit = int(spn_in_pgn_list[i - 1][1][0])
                            here_startbit = prev_startbit + prev_spn_len
                        here_tuple = list(spn_in_pgn_list[i])
                        here_tuple[1] = [here_startbit]
                        spn_in_pgn_list[i] = tuple(here_tuple)

                # update the maps
                pgn_object.update(
                    {
                        "SPNs": list(
                            map(operator.itemgetter(0), spn_in_pgn_list)
                        )
                    }
                )
                pgn_object.update(
                    {
                        "SPNStartBits": list(
                            map(operator.itemgetter(1), spn_in_pgn_list)
                        )
                    }
                )
                pgn_object.update(
                    {
                        "Temp_SPN_Order": list(
                            map(operator.itemgetter(2), spn_in_pgn_list)
                        )
                    }
                )

    @staticmethod
    def remove_underspecd_spns(j1939_pgn_db, j1939_spn_db):
        for pgn, pgn_object in j1939_pgn_db.items():
            spn_list = pgn_object.get("SPNs")
            if len(spn_list) > 1:
                spn_list = pgn_object.get("SPNs")
                spn_startbit_list = pgn_object.get("SPNStartBits")
                spn_order_list = pgn_object.get("Temp_SPN_Order")

                spn_in_pgn_list = zip(
                    spn_list, spn_startbit_list, spn_order_list
                )

                def should_remove(tup):
                    spn = tup[0]
                    spn_obj = j1939_spn_db.get(str(spn))
                    current_spn_length = spn_obj.get("SPNLength")
                    current_spn_delimiter = spn_obj.get("Delimiter")
                    if (
                        J1939daConverter.is_length_variable(current_spn_length)
                        and current_spn_delimiter is None
                    ):
                        print(
                            "Warning: removing SPN %s from PGN %s because it "
                            "is variable-length with no delimiter in a multi-SPN PGN. "
                            "This likely an under-specification in the DA."
                            % (spn, pgn),
                            file=sys.stderr,
                        )
                        return True
                    return False

                spn_in_pgn_list = [
                    tup for tup in spn_in_pgn_list if not should_remove(tup)
                ]

                # update the maps
                pgn_object.update(
                    {
                        "SPNs": list(
                            map(operator.itemgetter(0), spn_in_pgn_list)
                        )
                    }
                )
                pgn_object.update(
                    {
                        "SPNStartBits": list(
                            map(operator.itemgetter(1), spn_in_pgn_list)
                        )
                    }
                )
                pgn_object.update(
                    {
                        "Temp_SPN_Order": list(
                            map(operator.itemgetter(2), spn_in_pgn_list)
                        )
                    }
                )

    @staticmethod
    def sort_spns_by_order(j1939_pgn_db):
        for pgn, pgn_object in j1939_pgn_db.items():
            spn_list = pgn_object.get("SPNs")
            spn_startbit_list = pgn_object.get("SPNStartBits")
            spn_order_list = pgn_object.get("Temp_SPN_Order")

            spn_in_pgn_list = zip(spn_list, spn_startbit_list, spn_order_list)
            # sort numbers then letters
            spn_in_pgn_list = sorted(
                spn_in_pgn_list,
                key=lambda obj: (isinstance(obj[2], str), obj[2]),
            )

            # update the maps (now sorted by 'Temp_SPN_Order')
            pgn_object.update(
                {"SPNs": list(map(operator.itemgetter(0), spn_in_pgn_list))}
            )
            pgn_object.update(
                {
                    "SPNStartBits": list(
                        map(operator.itemgetter(1), spn_in_pgn_list)
                    )
                }
            )
            pgn_object.update(
                {
                    "Temp_SPN_Order": list(
                        map(operator.itemgetter(2), spn_in_pgn_list)
                    )
                }
            )

    @staticmethod
    def all_spns_positioned(spn_startbit_list):
        if len(spn_startbit_list) == 0:
            return True
        else:
            is_positioned = map(
                lambda spn_startbit: int(spn_startbit[0]) != -1,
                spn_startbit_list,
            )
            return functools.reduce(lambda a, b: a and b, is_positioned)

    def process_any_source_addresses_sheet(self, sheet):
        if self.j1939db.get("J1939SATabledb") is None:
            self.j1939db.update({"J1939SATabledb": OrderedDict()})
        j1939_sa_tabledb = self.j1939db.get("J1939SATabledb")

        header_row, header_row_num = self.get_header_row(sheet)

        source_address_id_col = self.get_any_header_column(
            header_row, "SOURCE_ADDRESS_ID"
        )
        name_col = self.get_any_header_column(header_row, "NAME")

        for i in range(header_row_num + 1, sheet.nrows):
            row = sheet.row_values(i)

            name = row[name_col]
            if name.startswith("thru") or name.startswith("through"):
                start_range = int(row[source_address_id_col])
                range_clues = name.replace("thru", "").replace("through", "")
                range_clues = range_clues.strip()
                end_range = int(range_clues.split(" ")[0])
                description = "".join(name.split(str(end_range))[1:]).strip()
                description = description + " " + row[name_col + 1]
                description = re.sub(r"^are ", "", description)
                description = description.strip()
                for val in range(start_range, end_range + 1):
                    j1939_sa_tabledb.update({str(val): description})
            else:
                val = str(int(row[source_address_id_col]))
                name = name.strip()
                j1939_sa_tabledb.update({val: name})
        return

    def convert(self, output_file):
        self.j1939db = OrderedDict()
        sheet_name = ["SPNs & PGNs", "SPs & PGs"]
        self.process_spns_and_pgns_tab(
            self.find_first_sheet_by_name(sheet_name)
        )
        sheet_name = "Global Source Addresses (B2)"
        self.process_any_source_addresses_sheet(
            self.find_first_sheet_by_name(sheet_name)
        )
        sheet_name = "IG1 Source Addresses (B3)"
        self.process_any_source_addresses_sheet(
            self.find_first_sheet_by_name(sheet_name)
        )

        out = open(output_file, "w") if output_file != "-" else sys.stdout

        try:
            out.write(json.dumps(self.j1939db, indent=2, sort_keys=False))
        except BrokenPipeError:
            pass

        if out is not sys.stdout:
            out.close()

        return

    def find_first_sheet_by_name(self, sheet_names):
        if not isinstance(sheet_names, list):
            sheet_names = [sheet_names]
        for sheet_name in sheet_names:
            for book in self.digital_annex_xls_list:
                if sheet_name in book.sheet_names():
                    sheet = book.sheet_by_name(sheet_name)
                    return sheet
        return None



# Copyright Andrew Dodd
import math
from binascii import hexlify
from numbers import Number
from typing import Any, Dict, Optional, Union

import attr

from decoda.exceptions import (
    CouldBeAnyLength,
    ErrorIndicatorRangeError,
    NotAvaiableRangeError,
    ParameterSpecificIndicatorError,
    UnknownReferenceError,
    UnsupportedLocationSpec,
)


@attr.s(frozen=True)
class Manufacturer:
    id = attr.ib()
    name = attr.ib()
    location = attr.ib()
    status = attr.ib()


@attr.s(frozen=True)
class BitLength:
    max_len = attr.ib()
    variable = attr.ib(default=False)

    def __attrs_post_init__(self):
        if (
            self.variable
            and self.max_len is not None
            and self.max_len % 8 != 0
        ):
            raise ValueError(
                "bit_length ({}) must be an 8-bit multiple".format(
                    self.max_len
                )
            )

    def could_fit_in_bytes(self, value):
        if self.variable:
            return True
        return self.max_len // 8 <= len(value)

    def expected_length(self, variable_pgn):
        if self.variable and variable_pgn:
            raise CouldBeAnyLength()
        return self

    def trim_to_length(self, val):
        if self.max_len is not None and val:
            max_bytes = self.max_len // 8
            val = val[:max_bytes]
        return val

    def is_equivalent_to(self, bits):
        return not self.variable and self.max_len == bits


def convert_to_int(i: Any) -> int:
    if isinstance(i, int):
        return i
    if isinstance(i, str) and i.startswith("0x"):
        return int(i[2:], 16)
    if isinstance(i, str) and i.startswith("0b"):
        return int(i[2:], 2)
    return int(i, 10)


def convert_encodings(value: Dict) -> Dict:
    encodings = {}
    for k, v in value.items():
        if isinstance(k, str) and "-" in k:
            start, end = k.split("-")
        else:
            start = k
            end = k

        start_int = convert_to_int(start)
        end_int = convert_to_int(end) + 1

        for k in range(start_int, end_int):
            encodings[k] = v
    return encodings


@attr.s(frozen=True)
class EncodedValue:
    encodings: Dict = attr.ib(converter=convert_encodings)
    bit_length = attr.ib()

    @encodings.validator
    def _check_encodings(self, attribute, value):
        if len(value) == 0:
            raise ValueError("must have at least 1 encoding")

    @bit_length.validator
    def _check_bit_length(self, attribute, value):
        if value.max_len is None or value.max_len < 1:
            raise ValueError("must have positive bit length")

    def decode_from_raw(
        self, value, start_spec, variable_pgn, prev_spn_ended, *args, **kwargs
    ):
        bit_len = self.bit_length.expected_length(variable_pgn)
        raw, end_byte_idx = extract_value_at_location(
            value, start_spec, bit_len
        )
        value = self.encodings.get(raw)
        if value:
            display_value = "{} ({})".format(value, raw)
        else:
            display_value = "No encoding ({})".format(raw)

        return {
            "raw": raw,
            "value": value,
            "display_value": display_value,
        }, end_byte_idx


@attr.s(frozen=True)
class ScalarValue:
    units = attr.ib()
    min = attr.ib()
    max = attr.ib()
    offset = attr.ib()
    scale = attr.ib()
    bit_length = attr.ib()

    @bit_length.validator
    def _check_bit_length(self, attribute, value):
        if value.max_len is None or value.max_len < 1:
            raise ValueError("must have positive bit length")

    def __attrs_post_init__(self):
        if self.min and self.max and not self.min < self.max:
            raise ValueError("min must be less than max")

    def as_display_value(self, value):
        return "{} {}".format(value, self.units) if self.units else str(value)

    def check_in_valid_range(self, raw):
        not_available_thres = 0xFF
        error_ind_thres = 0xFE
        param_specific_thres = 0xFB
        format_offset = lambda val, thres: val - thres

        if self.bit_length.is_equivalent_to(8):
            format_offset = lambda val, thres: None
        elif self.bit_length.is_equivalent_to(16):
            not_available_thres <<= 8
            error_ind_thres <<= 8
            param_specific_thres <<= 8
        elif self.bit_length.is_equivalent_to(32):
            not_available_thres <<= 24
            error_ind_thres <<= 24
            param_specific_thres <<= 24
        else:
            return

        if raw >= not_available_thres:
            raise NotAvaiableRangeError(
                format_offset(raw, not_available_thres)
            )
        if raw >= error_ind_thres:
            raise ErrorIndicatorRangeError(format_offset(raw, error_ind_thres))
        if raw >= param_specific_thres:
            raise ParameterSpecificIndicatorError(
                format_offset(raw, param_specific_thres)
            )

    def decode_from_raw(
        self,
        value,
        start_spec,
        variable_pgn,
        prev_spn_ended,
        ignore_range=False,
        **kwargs,
    ):
        bit_len = self.bit_length.expected_length(variable_pgn)
        raw, end_byte_idx = extract_value_at_location(
            value, start_spec, bit_len
        )

        try:
            self.check_in_valid_range(raw)

            value = raw
            if self.scale:
                value *= self.scale

            if self.offset:
                value += self.offset

            display_value = self.as_display_value(value)

            if not ignore_range:
                if self.min and value < self.min:
                    display_value = (
                        "{} (encoded value {} was clipped to min)".format(
                            self.as_display_value(self.min),
                            self.as_display_value(value),
                        )
                    )
                    value = self.min
                if self.max and value > self.max:
                    display_value = (
                        "{} (encoded value {} was clipped to max)".format(
                            self.as_display_value(self.max),
                            self.as_display_value(value),
                        )
                    )
                    value = self.max

        except (
            NotAvaiableRangeError,
            ErrorIndicatorRangeError,
            ParameterSpecificIndicatorError,
        ) as e:
            value = e.value()
            display_value = e.display_value()

        return {
            "raw": raw,
            "value": value,
            "display_value": display_value,
        }, end_byte_idx


def convert_if_hex(value: str) -> Optional[str]:
    if value is None:
        return value
    if value.startswith("0x"):
        return chr(int(value, 16))
    return value


@attr.s(frozen=True)
class TextValue:
    bit_length = attr.ib()
    delimiter: Optional[str] = attr.ib(default=None, converter=convert_if_hex)
    # I.e. another SPN that specifies its length
    length_spn = attr.ib(default=None)

    def __attrs_post_init__(self):
        if (
            self.bit_length.variable
            and self.delimiter is None
            and self.length_spn is None
        ):
            raise ValueError(
                "Variable length SPN must specify delimiter or length SPN"
            )

    def decode_from_raw(
        self,
        value,
        start_spec,
        variable_pgn,
        prev_spn_ended,
        already_decoded=None,
        **kwargs,
    ):
        start_idx = self._get_start(start_spec, prev_spn_ended)
        val = str(value[start_idx:], "ascii")

        # Hmm... should deal with not variable length?
        if not variable_pgn and not self.bit_length.could_fit_in_bytes(value):
            raise ValueError("insufficent bytes")

        tightly_packed_strings = False
        encoded_length = None
        if self.length_spn:
            try:
                decoded_spn = next(
                    decoded
                    for decoded in already_decoded
                    if decoded.id == self.length_spn
                )
            except StopIteration:
                raise ValueError(
                    f"Length SPN: {self.length_spn} not found in already_decoded"
                )
            encoded_length = decoded_spn.value
            tightly_packed_strings = True

        if encoded_length:
            val = val[:encoded_length]
        elif self.delimiter:
            val = val.split(self.delimiter)[0]

        val = self.bit_length.trim_to_length(val)
        end_byte_idx = start_idx + len(val)

        if tightly_packed_strings:
            end_byte_idx -= 1

        return {"raw": val, "value": val, "display_value": val}, end_byte_idx

    def _get_start(self, start_spec, prev_spn_ended):
        start = start_spec.split()[0].split("-")[0]
        try:
            return int(start) - 1
        except ValueError:
            return prev_spn_ended + 1


@attr.s(frozen=True)
class ByteArrayValue:
    bit_length = attr.ib()

    def decode_from_raw(
        self, value, start_spec, variable_pgn, prev_spn_ended, *args, **kwargs
    ):
        start = start_spec.split()[0].split("-")[0]
        try:
            start_idx = int(start) - 1
        except ValueError:
            start_idx = prev_spn_ended + 1

        val = value[start_idx:]
        end_byte_idx = start_idx + len(val)

        return {
            "raw": val,
            "value": val,
            "display_value": val.hex(),
        }, end_byte_idx


@attr.s(frozen=True)
class CustomFunctionValue:
    custom = attr.ib()
    bit_length = attr.ib()
    non_custom_alternative = attr.ib()
    custom_args = attr.ib(default=None)

    def decode_from_raw(
        self,
        value,
        start_spec,
        variable_pgn,
        prev_spn_ended,
        *args,
        already_decoded=None,
        **kwargs,
    ):
        bit_len = self.bit_length.expected_length(variable_pgn)
        raw, end_byte_idx = extract_value_at_location(
            value, start_spec, bit_len
        )

        val = self.custom(
            value=raw,
            already_decoded=already_decoded,
            custom_args=self.custom_args,
            non_custom_alternative=self.non_custom_alternative,
            start_spec=start_spec,
            variable_pgn=variable_pgn,
            prev_spn_ended=prev_spn_ended,
            original_value=value,
        )
        if val is None:
            return None, prev_spn_ended

        return {
            "raw": raw,
            "value": val,
            "display_value": f"{val} ({raw})",
        }, end_byte_idx


@attr.s(frozen=True)
class SPNFields:
    id = attr.ib()
    name = attr.ib()
    description = attr.ib()


@attr.s(frozen=True)
class SPN(SPNFields):
    value_decoder = attr.ib()

    def decode(
        self,
        value,
        start_spec,
        variable_pgn,
        prev_spn_ended_idx,
        *args,
        **kwargs,
    ):
        decoded, end_byte_idx = self.value_decoder.decode_from_raw(
            value,
            start_spec,
            variable_pgn,
            prev_spn_ended_idx,
            *args,
            **kwargs,
        )
        if decoded is None:
            return None, prev_spn_ended_idx
        return DecodedSPN.build(self, **decoded), end_byte_idx


@attr.s(frozen=True)
class DecodedSPN(SPNFields):
    raw: Number = attr.ib()
    value: Union[Number, str] = attr.ib()
    display_value: str = attr.ib()

    @classmethod
    def build(
        cls,
        spn: SPN,
        raw: Number,
        value: Union[Number, str],
        display_value: str,
    ):
        return cls(
            id=spn.id,
            name=spn.name,
            description=spn.description,
            raw=raw,
            value=value,
            display_value=display_value,
        )

    def __str__(self):
        return f"SPN:{self.id}: {self.name} = {self.value}"


def order_start_positions(start_pos_spec):
    if "to" in start_pos_spec:
        first = start_pos_spec.split("to")[0]
    elif "," in start_pos_spec:
        return order_start_positions(start_pos_spec.split(",")[0])
    else:
        first = start_pos_spec.split("-")[0]

    if first.startswith("0x"):
        raise UnsupportedLocationSpec("We don't support this yet")
    try:
        return float(first)
    except ValueError:
        return first.lower()


# We need this because our decoding relies on the order of the records
class OrderingRecord:
    def __init__(self, start, spn):
        self.start = start
        self.spn = spn

    def __repr__(self):
        return "OrderingRecord({}, {})".format(self.start, self.spn)

    def __lt__(self, other):
        if other.__class__ is not self.__class__:
            return NotImplemented

        own_start = self.__first_element_of_start_position()
        other_start = other.__first_element_of_start_position()

        # If the they are comparable, order them
        if type(own_start) == type(other_start):
            return own_start < other_start

        # if they are not...then order "floats" first
        return isinstance(own_start, float)

    def __first_element_of_start_position(self):
        return order_start_positions(self.start)


def byte_and_bit_from_num(num):
    parts = num.split(".")
    byte = parts[0]
    bit = parts[1] if len(parts) > 1 else 1
    return int(byte), int(bit)


def mask(length):
    return (1 << length) - 1


def extract_value_at_location(payload, location, bit_length):
    """
    Extract a "numeric" payload from the byte array at the provided location

    Returns: (value, last_byte_idx)

    where
     - value: the "bit_length" at "location" from the payload, and
     - last_byte_idx: is the index of the last byte that the value was in.
    """
    sections = [s.strip() for s in location.split(",")]
    if len(sections) > 1:
        if len(sections) > 2:
            raise UnsupportedLocationSpec(
                "Should not have more than 2 sections in location spec"
            )
        contiguous, extra = sections
        if "-" in extra:
            raise UnsupportedLocationSpec(
                "Should not have a range specified in non-continuous part of location spec"
            )

        # assume that the contigous section is multiple bytes in length
        contiguous_length = 8
        if "-" in contiguous:
            # we have a multi byte contiguous section
            contiguous_start, contiguous_end = (
                int(x) for x in contiguous.split("-")
            )
            contiguous_length = (1 + contiguous_end - contiguous_start) * 8
        elif "." in contiguous:
            # It appears to be a new spec?
            contiguous_length = 9 - int(contiguous.split(".")[1])

        contiguous_value, _ = extract_value_at_location(
            payload, contiguous, BitLength(contiguous_length)
        )
        extra_value, last_byte_idx = extract_value_at_location(
            payload, extra, BitLength(bit_length.max_len - contiguous_length)
        )
        extra_value <<= contiguous_length

        return extra_value + contiguous_value, last_byte_idx

    section = sections[0]
    if "-" not in section:
        # single byte
        byte, bit = byte_and_bit_from_num(section)
        byte -= 1
        bit -= 1
        payload = payload[byte:]

        if not bit_length.could_fit_in_bytes(payload):
            raise ValueError(
                "not enough bits for SPN - need {}, from {}.{} in payload[{}]".format(
                    bit_length.max_len, byte, byte, hexlify(payload)
                )
            )

        v = int.from_bytes(payload, byteorder="little")
        v >>= bit
        v &= mask(bit_length.max_len)

        additional_bytes = math.ceil((bit + bit_length.max_len) / 8.0) - 1
        last_byte_idx = byte + additional_bytes
        return v, last_byte_idx

    # multi byte section
    first, last = section.split("-")
    f_byte, f_bit = byte_and_bit_from_num(first)
    l_byte, l_bit = byte_and_bit_from_num(last)

    if l_bit != 1:
        raise UnsupportedLocationSpec(
            "Should not have a non-first-bit aligned location spec for contiguous bytes"
        )

    # last address starts on a byte boundary...so everything should be
    # contiguous???
    return extract_value_at_location(payload, first, bit_length)


@attr.s(frozen=True)
class PGN:
    id = attr.ib()
    name = attr.ib()
    description = attr.ib()
    transmission_rate = attr.ib()
    length = attr.ib()
    ordering_records = attr.ib()
    repeatable_spns = attr.ib(factory=list)
    acronym = attr.ib(default=None)

    def is_repeatable(self, spn):
        return spn.id in self.repeatable_spns

    def decode(self, value):
        decoded = []
        pgn_is_variable = self.length.variable
        ended_in_byte_idx = -1
        repeatable_spn_idx = None

        for idx, record in enumerate(self.ordering_records):
            start, spn = record.start, record.spn

            if self.is_repeatable(spn):
                repeatable_spn_idx = idx
                break
            else:
                # need to get the byte that the result came from???
                result, ended_in_byte_idx = spn.decode(
                    value,
                    start,
                    pgn_is_variable,
                    ended_in_byte_idx,
                    already_decoded=decoded,
                )
                if result is not None:
                    decoded.append(result)

        if repeatable_spn_idx:
            # we found the first repeatable spn above, so came here
            # The remaining are the repeatable ones
            # we assume they will all be the same length...(I hope!)
            repeatables = self.ordering_records[repeatable_spn_idx:]

            # Do the first set of repeatable SPNs
            start_idx = ended_in_byte_idx
            for record in repeatables:
                start, spn = record.start, record.spn
                result, ended_in_byte_idx = spn.decode(
                    value,
                    start,
                    pgn_is_variable,
                    ended_in_byte_idx,
                    already_decoded=decoded,
                )
                decoded.append(result)
            end_idx = ended_in_byte_idx

            # Now we have the length of the repeatable section in bytes
            # trim the value from the front and keep doing it until we
            # run out of bytes
            bytes_in_repeat = end_idx - start_idx
            while len(value) > end_idx + bytes_in_repeat:
                value = value[bytes_in_repeat:]
                for record in repeatables:
                    start, spn = record.start, record.spn
                    result, ended_in_byte_idx = spn.decode(
                        value,
                        start,
                        pgn_is_variable,
                        ended_in_byte_idx,
                        already_decoded=decoded,
                    )
                    decoded.append(result)

        return decoded


@attr.s(frozen=True)
class Address:
    num = attr.ib()

    def is_broadcast(self):
        return self.num == -1

    @num.validator
    def _check_num(self, attribute, value):
        if not -1 <= value <= 255:
            raise ValueError("address out of range")

    @staticmethod
    def from_pf_and_ps(pf, ps):
        if pf >= 0xF0:
            return BroadcastAddress

        return Address(ps)


BroadcastAddress = Address(-1)


@attr.s(frozen=True)
class IndustryGroup:
    id = attr.ib()
    description = attr.ib()
    preferred_addresses = attr.ib()

    def preferred_address_name(self, id):
        return self.preferred_addresses.get(id)
